#include	"../math/mymath.h"
#include	"stdio.h"
#include	"Luna2.h"



// constructor
Luna2_VL::Luna2_VL(OBJHANDLE hObj, int fmodel): VESSEL2M(hObj, fmodel)
{
	// VESSEL2M crash parameters
	VvCrash = DRand(10, 50); 
	VhCrash = DRand(30, 60);
	VMaxBlast = DRand(100, 1000); 
	QMax = DRand(10000, 50000);

	// mesh
	int Mesh1 = AddMesh("r7_VL\\Luna2");

	// define animation and its defaults
	timer = 0;						// initial animations delay timer
	bAnimate = false;		
	CurrentAnimation = AT_NONE;
	tempAnimationPercent = 0;
	DefineAnimations();

	// start with all animations in the initial states
	AntennaRotationPercent = 0;
	TapeExtensionPercent = 0;
	TapeRotationPercent = 0;

	// moon reference
	hMoon = oapiGetGbodyByName ("Moon");

	// pennants
	bAllowFocus = true;
	bSwitchFocus = false;
}


void Luna2_VL::clbkSetClassCaps(FILEHANDLE cfg)
{
	// physics and geometry
	SetSize(1.0);
	SetPMI(PMI);
	SetCrossSections(CS);
	SetCW (0.2, 0.2, 0.2, 0.2); 
	SetRotDrag (_V(0.1, 0.1, 0.1));
	SetLiftCoeffFunc(0);
	SetPitchMomentScale(1e-6);
	SetBankMomentScale(1e-6);
	SetEmptyMass(EMPTY_MASS);
	SetSurfaceFrictionCoeff(1e5, 1e5);
	SetTouchdownPoints (_V( 0, 1, 2), _V( 2, -1, -2), _V(-2, -2, -2));

	// dummy (but full) propellant tank
	CreatePropellantResource(EXPLOSIVE_MASS, EXPLOSIVE_MASS);
}


void Luna2_VL::clbkSaveState(FILEHANDLE scn)
{
	// do inherited VESSEL2M stuff before writing custom stuff
	VESSEL2M::clbkSaveState (scn);
	oapiWriteScenario_string (scn, "=========== LUNA-2 vars", "");

	if (timer > 0)
		oapiWriteScenario_float(scn, "TIMER", timer);

	// animating right now or animation finished!
	if (CurrentAnimation > AT_NONE)
	{
		// which animation is current/last?
		oapiWriteScenario_int (scn, "ANTENNAS_PHASE", CurrentAnimation);

		if (CurrentAnimation == AT_ANTENNA)
			oapiWriteScenario_float (scn, "ANTENNAS_STATUS", AntennaRotationPercent);
		else if (CurrentAnimation == AT_TAPE_EXT)
			oapiWriteScenario_float (scn, "ANTENNAS_STATUS", TapeExtensionPercent);
		else if (CurrentAnimation == AT_TAPE_ROT)
			oapiWriteScenario_float (scn, "ANTENNAS_STATUS", TapeRotationPercent);
	}

	// pennants focusing
	if (!bAllowFocus)
		oapiWriteScenario_int (scn, "ALLOW_PENNATS_FOCUS", bAllowFocus);
	if (bSwitchFocus)
		oapiWriteScenario_int (scn, "SWITCH_PENNATS_FOCUS", bSwitchFocus);
}


void Luna2_VL::ParseScenarioLineEx(char *line, void *status)
{
	if (!strnicmp (line, "TIMER", 5))
		sscanf (line+5, "%lf", &timer);

	// antennas state or incomplete animation

	else if (!strnicmp (line, "ANTENNAS_PHASE", 14))
		sscanf (line+14, "%d", &CurrentAnimation);

	else if (!strnicmp (line, "ANTENNAS_STATUS", 15))
		sscanf (line+15, "%lf", &tempAnimationPercent);

	// pennants focusing
	else if (!strnicmp (line, "ALLOW_PENNATS_FOCUS", 19))
		sscanf (line+19, "%d", &bAllowFocus);

	else if (!strnicmp (line, "SWITCH_TO_PENNAT", 19))
		sscanf (line+19, "%d", &bSwitchFocus);

	else 
		VESSEL2M::ParseScenarioLineEx(line, status);
}


// called after all vessel in scn are created, but before first timestep
void Luna2_VL::FinishConfiguration()
{
	// update animation percents
	switch (CurrentAnimation)
	{
		case AT_TAPE_ROT :
		{
			TapeRotationPercent = tempAnimationPercent;
			TapeExtensionPercent = 1;
			AntennaRotationPercent = 1;
			break;
		}
		
		case AT_TAPE_EXT :
		{
			TapeExtensionPercent = tempAnimationPercent;
			AntennaRotationPercent = 1;
			break;
		}

		case AT_ANTENNA :
		{
			AntennaRotationPercent = tempAnimationPercent;
			break;
		}
	}

	// fast-forward animations
	SetAnimation(AntennaRotation, AntennaRotationPercent);
	SetAnimation(TapeExtension, TapeExtensionPercent);
	SetAnimation(TapeRotation, TapeRotationPercent);

	// is animation still running?
	bAnimate = ( (tempAnimationPercent > 0) && (tempAnimationPercent < 1) );

	//VESSEL2M stuff
	VESSEL2M::FinishConfiguration();
}


int Luna2_VL::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{
	// only process keydown events
	if (!down) 
		return 0; 

	// shift combinations are reserved
	if (KEYMOD_SHIFT (kstate)) 
		return 0; 


	// Control combinations!
	if (KEYMOD_CONTROL (kstate))
	{
		// switch focus to one of the pennats after impact? (only if focusing is allowed)
		if ( (key == OAPI_KEY_J) && (bAllowFocus) ) 
		{
			bSwitchFocus = !bSwitchFocus;

			if (bSwitchFocus)
				ShowAnnotationMessage("Switch focus to pennant after impact? YES!");
			else
				ShowAnnotationMessage("Switch focus to pennant after impact? NO!");
		}

		// allow focus on pennants
		if ( (key == OAPI_KEY_F) )
		{
			bAllowFocus = !bAllowFocus;

			if (bAllowFocus)
				ShowAnnotationMessage("Allow pennants to be focused after impact? YES!");
			else
			{
				ShowAnnotationMessage("Allow pennants to be focused after impact? NO!");
				bSwitchFocus = false;
			}
		}

		return 0; 
	}

	return 0;
}


void Luna2_VL::clbkPostStep(double simt, double SimDT, double mjd)
{
	// common processing for all VESSEL2M-derived objects
	TimestepForVESSEL2M();

	SimDt = SimDT;

	// are we still attached to the rocket? 
	if (bAttached)
	{
		// this indicates our separation from the rocket
		if (GetAttachmentStatus(GetAttachmentHandle(TOPARENT, 0)) == NULL)
		{
			// pyro sound
			PlayVesselWave3(SoundLibID, SND_PYRO);
			
			// start antennas timer
			timer = 10;

			bAttached = false;
		}

		return;
	}

	// update antenna timer
	if (timer > 0)
	{
		timer -= SimDT;
		if (timer <= 0)
			AntennaRotationStart();
	}

	// run animation
	if (bAnimate)
		Animate();

	// the rest runs only if we are inside moon's sphere of influence
	if ( GetSurfaceRef() == hMoon )
	{
		// limit time acceleration depending on the moon's proximity
		double Alt = GetAltitude();
		double MaxTimeAccel;

		if (Alt < 30000)
			MaxTimeAccel = 1;
		else if (Alt < 300000)
			MaxTimeAccel = 10;
		else if (Alt < 3000000)
			MaxTimeAccel = 100;
		else if (Alt < 30000000)
			MaxTimeAccel = 1000;
		else
			MaxTimeAccel = 10000;

		if (oapiGetTimeAcceleration() > MaxTimeAccel)
			oapiSetTimeAcceleration(MaxTimeAccel);
	}
}


void Luna2_VL::Explode()
{
	void *hPennant;
	VESSEL *VesselIntf;
	int TargetPennant;

	// prepare to switch focus
	if (bSwitchFocus)
		TargetPennant = (int)DRand(0, 50);
	else
		TargetPennant = -1;

	// pennant name
	char Name[100];

	// clone pennants' status from the current one
	VESSELSTATUS vs;
	GetStatus(vs);

	//	create pennants
	for (int i=0; i<50;i++)
	{
		// prepare a name for a pennant object 
		sprintf(Name, "%s-Pennant-%d", GetName(), i);

		const double PND_VMIN = -500.; 
		const double PND_VMAX = -PND_VMIN; 
		const double PND_VSPAN = PND_VMAX - PND_VMIN; 

		const double PND_OMIN = -10*RAD; 
		const double PND_OMAX = -PND_OMIN; 
		const double PND_OSPAN = PND_OMAX - PND_OMIN; 

		// update-randomize pennant's status
		vs.rvel.x += DRand(PND_VMIN, PND_VSPAN);
		vs.rvel.y += DRand(PND_VMIN, PND_VSPAN);
		vs.rvel.z += DRand(PND_VMIN, PND_VSPAN);

		vs.vrot.x += DRand(PND_OMIN, PND_OSPAN);
		vs.vrot.y += DRand(PND_OMIN, PND_OSPAN);
		vs.vrot.z += DRand(PND_OMIN, PND_OSPAN);


		// create a pennant object 
		hPennant = oapiCreateVessel(Name, "r7_VL\\Luna2Pennant", vs); 

		// update pennant object focus
		if (!bAllowFocus)
		{
			VesselIntf = oapiGetVesselInterface(hPennant);
			VesselIntf->SetEnableFocus(0);
		}

		// switch focus
		if ( (i == TargetPennant) && (oapiGetFocusObject() == GetHandle()) )
			oapiSetFocusObject(hPennant);
	}

	//VESSEL2M stuff
	VESSEL2M::Explode();
}


	// ===== ANIMATIONS =====


void Luna2_VL::DefineAnimations()
{
	// Antennas
	MGROUP_ROTATE *Antennas[4];

	static UINT grAntenna0[2] = {8, 7}; 
	static UINT grAntenna1[2] = {9, 10}; 
	static UINT grAntenna2[2] = {5, 6}; 
	static UINT grAntenna3[2] = {2, 11}; 

	Antennas[0] = new MGROUP_ROTATE(0, grAntenna0, 2, ANTEN00_REF, ANTEN00_AXIS,  ANTENNA_ROTATION_ANGLE);
	Antennas[1] = new MGROUP_ROTATE(0, grAntenna1, 2, ANTEN01_REF, ANTEN01_AXIS, -ANTENNA_ROTATION_ANGLE);
	Antennas[2] = new MGROUP_ROTATE(0, grAntenna2, 2, ANTEN02_REF, ANTEN02_AXIS,  ANTENNA_ROTATION_ANGLE);
	Antennas[3] = new MGROUP_ROTATE(0, grAntenna3, 2, ANTEN03_REF, ANTEN03_AXIS, -ANTENNA_ROTATION_ANGLE);

	AntennaRotation = CreateAnimation (0);

	AddAnimationComponent (AntennaRotation, 0, 1, Antennas[0]);
	AddAnimationComponent (AntennaRotation, 0, 1, Antennas[1]);
	AddAnimationComponent (AntennaRotation, 0, 1, Antennas[2]);
	AddAnimationComponent (AntennaRotation, 0, 1, Antennas[3]);

	// Tapes 
	static UINT grTapes1[1] = {17}; 
	static UINT grTapes2[1] = {15}; 

	// tapes extesion
	MGROUP_SCALE *TapesScale[2];

	TapesScale[0] = new MGROUP_SCALE(0, grTapes1, 1, _V( 0.146, 0, -0.414), _V(1, 1, 20));
	TapesScale[1] = new MGROUP_SCALE(0, grTapes2, 1, _V(-0.146, 0, -0.414), _V(1, 1, 20));

	TapeExtension = CreateAnimation (0);

	AddAnimationComponent (TapeExtension, 0, 1, TapesScale[0]);
	AddAnimationComponent (TapeExtension, 0, 1, TapesScale[1]);

	// tapes rotation
	MGROUP_ROTATE *TapesRot[2];

	TapesRot[0] = new MGROUP_ROTATE(0, grTapes1, 1, _V( 0.146, 0, -0.414), _V(0, 1, 0), (float)(-45*RAD));
	TapesRot[1] = new MGROUP_ROTATE(0, grTapes2, 1, _V(-0.146, 0, -0.414), _V(0, 1, 0), (float) (45*RAD));

	TapeRotation = CreateAnimation (0);

	AddAnimationComponent (TapeRotation, 0, 1, TapesRot[0]);
	AddAnimationComponent (TapeRotation, 0, 1, TapesRot[1]);
}



void Luna2_VL::Animate()
{
	if (CurrentAnimation == AT_ANTENNA)
		AnimateAntennaRotation();
	
	else if (CurrentAnimation == AT_TAPE_EXT)
		AnimateTapeExtension();

	else if (CurrentAnimation == AT_TAPE_ROT)
		AnimateTapeRotation();
}


void Luna2_VL::AnimateAntennaRotation()
{
	double dPercent = ANTENNA_ROTATION_SPEED * SimDt;

	if (AntennaRotationPercent < 1.0)
		AntennaRotationPercent = min (1.0, AntennaRotationPercent + dPercent);
	else
		AntennaRotationEnd();

	SetAnimation (AntennaRotation, AntennaRotationPercent);
}


void Luna2_VL::AnimateTapeExtension()
{
	double dPercent = TAPE_EXTENSION_SPEED * SimDt;

	if (TapeExtensionPercent < 1.0)
		TapeExtensionPercent = min (1.0, TapeExtensionPercent + dPercent);
	else
		TapeExtensionEnd();

	SetAnimation (TapeExtension, TapeExtensionPercent);
}


void Luna2_VL::AnimateTapeRotation()
{
	double dPercent = TAPE_ROTATION_SPEED * SimDt;

	if (TapeRotationPercent < 1.0)
		TapeRotationPercent = min (1.0, TapeRotationPercent + dPercent);
	else
		TapeRotationEnd();

	SetAnimation (TapeRotation, TapeRotationPercent);
}


void Luna2_VL::AntennaRotationStart()
{
	// ignore multipe restarts
	if (bAnimate)
		return;

	CurrentAnimation = AT_ANTENNA;

	// pyro sound
	PlayVesselWave3(SoundLibID, SND_PYRO);

	// finally, allow animation
	bAnimate = true;
}


// same as TapeExtensionStart()
void Luna2_VL::AntennaRotationEnd()
{
	PlayVesselWave3(SoundLibID, SND_CLANG);

	CurrentAnimation = AT_TAPE_EXT;

	PlayVesselWave3(SoundLibID, SND_MOTOR, LOOP);
}


// same as TapeRotationStart()
void Luna2_VL::TapeExtensionEnd()
{
	StopVesselWave3(SoundLibID, SND_MOTOR);

	CurrentAnimation = AT_TAPE_ROT;
}


void Luna2_VL::TapeRotationEnd()
{
	bAnimate = false;
}



		// SOUND
void Luna2_VL::clbkPostCreation()
{
	SoundLibID = ConnectToOrbiterSoundDLL3(GetHandle());

	SoundOptionOnOff3(SoundLibID, PLAYMAINTHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYATTITUDETHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYLANDINGANDGROUNDSOUND, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAIRSPEED, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAMBIANCEWHENLANDED, true);

	SoundOptionOnOff3(SoundLibID, PLAYCOUNTDOWNWHENTAKEOFF, false);
	SoundOptionOnOff3(SoundLibID, PLAYWHENATTITUDEMODECHANGE, false);
	SoundOptionOnOff3(SoundLibID, PLAYGPWS, false);
	SoundOptionOnOff3(SoundLibID, PLAYHOVERTHRUST, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKINGSOUND, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADARBIP, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKLANDCLEARANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINAIRCONDITIONING, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINRANDOMAMBIANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADIOATC, false);
	SoundOptionOnOff3(SoundLibID, DISPLAYTIMER, false);

	RequestLoadVesselWave3(SoundLibID, SND_PYRO, "Sound\\r7_VL\\Pyro.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_CLANG, "Sound\\r7_VL\\Clang.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_MOTOR, "Sound\\r7_VL\\Motor.wav", BOTHVIEW_FADED_FAR);
}




DLLCLBK VESSEL *ovcInit(OBJHANDLE hvessel, int flightmodel)
{
	return new Luna2_VL(hvessel, flightmodel);
}

DLLCLBK void ovcExit(VESSEL *vessel)
{
	if (vessel) delete (Luna2_VL*) vessel;
}


