#include "..\r7_vessel2m_VL\vessel2m.h"



// geometry and physics 
const		double		EMPTY_MASS = 258.5;  // full mass 278.5
const		VECTOR3		PMI = {0.25, 0.25, 0.25};
const		VECTOR3		CS = {1.86, 1.86, 1.86};

// antenna rotation speed
const double ANTENNA_ROTATION_SPEED = 70*RAD;  // rad/s
const double TAPE_EXTENSION_SPEED = 70*RAD;  
const double TAPE_ROTATION_SPEED = 70*RAD;  
const double COVER_ROTATION_SPEED = 10*RAD;

const float ANTENNA_ROTATION_ANGLE = 85*(float)PI/180;

// antennas rotation centers
const		VECTOR3		ANTEN00_REF = { 0.358, 0, 0.404};
const		VECTOR3		ANTEN01_REF = {0, -0.358, 0.404};
const		VECTOR3		ANTEN02_REF = {-0.358, 0, 0.404};
const		VECTOR3		ANTEN03_REF = { 0, 0.358, 0.404};

// antennas rotation axis
const		VECTOR3		ANTEN00_AXIS = { 0, 1, 0};
const		VECTOR3		ANTEN01_AXIS = {-1, 0, 0};
const		VECTOR3		ANTEN02_AXIS = {0, -1, 0};
const		VECTOR3		ANTEN03_AXIS = { 1, 0, 0};

// sound
const int SND_PYRO = 1;
const int SND_CLANG = 2;
const int SND_PHOTO = 3;
const int SND_MOTOR = 4;
const int SND_RADIO = 5;

// radio transmission
const double ALT_TRANSM_MIN = 20000000; // transmission range by altitude, m
const double ALT_TRANSM_MAX = 50000000;
const double ELEV_TRANSM_MIN = 20; // deg
const double LAT_EUP = 45.189090 * RAD; // Eupatoria coordinates
const double LON_EUP = 33.18689 * RAD;
const double RADIUS0 = 6.37101e6; // Earth radius

// gas RCS

const double GAS_RCS_FUEL_MASS = 20;
const double GAS_RCS_ISP = 100;
const double THRUST_GAS_RSC = 5;
const double GAS_RCS_EXH_LENGTH = 0.15;
const double GAS_RCS_EXH_WIDTH = 0.004;

// center of mount point 
const double GAS_X = 0.276;
const double GAS_Z = -0.444;

const double D_GAS_X = 0.012;
const double D_GAS_Y = 0.012;
const double D_GAS_Z = 0.012;

const double GAS_YAW_X_L_0 = -GAS_X - D_GAS_X;
const double GAS_YAW_X_L_1 =  GAS_X - D_GAS_X;
const double GAS_YAW_X_R_0 = -GAS_X + D_GAS_X;
const double GAS_YAW_X_R_1 =  GAS_X + D_GAS_X;

const double GAS_YAW_Z_L_0 =  GAS_Z + D_GAS_Z;
const double GAS_YAW_Z_L_1 =  GAS_Z - D_GAS_Z;
const double GAS_YAW_Z_R_0 =  GAS_Z - D_GAS_Z;
const double GAS_YAW_Z_R_1 =  GAS_Z + D_GAS_Z;

const double GAS_YAW_ANGLE_X = 45*RAD;
const double GAS_YAW_ANGLE_Z = 90*RAD - GAS_YAW_ANGLE_X;

// photo
const int MAX_PHOTO_COUNT = 11; // 0 to 10, 0 is reserved for test table and 10 film frames can be used for photos
#define LOADBMP(id) ( (HBITMAP) LoadImage(NULL, id, IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE|LR_CREATEDIBSECTION) )

// photo panel areas
const int PHOTO_AREA = 0;
const int NUMBER_AREA = 1;
const int BTN_LEFT	= 2;
const int BTN_RIGHT	= 3;
const int BTN_TR = 4;
const int SIGNAL_AREA = 5;




class Luna3_VL: public VESSEL2M
{
public:
	Luna3_VL(OBJHANDLE, int);

	// callbacks
	void clbkSetClassCaps(FILEHANDLE);
	void clbkPostCreation();
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	int clbkConsumeDirectKey (char *keystate);
	void clbkPostStep(double, double, double);
	void clbkNavMode (int mode, bool active);
	bool clbkLoadGenericCockpit ();
	bool clbkLoadPanel (int id);
	bool clbkPanelRedrawEvent (int id, int event, SURFHANDLE surf);
	bool clbkPanelMouseEvent (int id, int event, int mx, int my);

	// VESSEL2M overrides
	void clbkSaveState (FILEHANDLE scn);
	void ParseScenarioLineEx(char *line, void *status);
	void FinishConfiguration ();
    void GetVesselMessage(VesselMessage VM);

private:
	// stored and running vars
	OBJHANDLE hMoon;			//+--
	OBJHANDLE hEarth;			//+--
	double Alt;					//---
	double MaxTimeAccel;		//+--
	double PastAlt;				//+--
	VECTOR3 xplus;				//+--

	// RCS, gas
	PROPELLANT_HANDLE GasRCSTank;						//---

	// timestep processing functions by flight stage
    enum chk_do_types 
	{ 
		CDT_DETACHED,	// are we still attached to parent object (block E)?
		CDT_MOONINFL,	// did we enter Moon influence spere? 
		CDT_MOONPERI,	// did we reach Moon pericenter? also, consider very low approach and/or hit for time accel purposes...
		CDT_PHOTOSTART,	// did we reach photo start altitude?
		CDT_PHOTOEND,	// did we reach photo end altitude?
		CDT_EARTHINFL,	// did we enter Earth influence spere? 
		CDT_TRANSSTART,	// did we reach photo transmission altitude?
		CDT_TRANSEND,	// did we reach photo transmission altitude?
		CDT_INERT		// done! 
	} 
	ChkDoType;	//+++ current CheckDo type for saving-restoring

	void CurrentConditionCheck();

	void CheckDetached();
	void DoDetach();

	void CheckMoonInfluence();
	void DoEnterMoonInfluence();

	void CheckMoonPericenter();
	void DoPassMoonPericenter();

	void CheckPhotoStart();
	void DoEnterPhotoWindow();

	void CheckPhotoEnd();
	void DoLeavePhotoWindow();

	void CheckEarthInfluence();
	void DoEnterEarthInfluence();

	void CheckTransmissionStart();
	void DoEnterTransmissionWindow();

	void CheckTransmissionEnd();
	void DoLeaveTransmissionWindow();

	// flight program autopilot
	bool bManualControl;						//+++
	void ToggleManualAndAutoplilot();
	void SetAutopilotControl();


	// moon tracking autopilot
	bool bTrackMoon;				//+++
	bool bApKillRot;				//+??
	bool bApKillRotStopped;			//+--
	VECTOR3 CurrentOrientLevels;	//--- xyz are pitch-yaw-roll signals in 0+-1 range.
	VECTOR3 omegas;					//--- angular velocities in the timestep, dg/s
	VECTOR3 OmegaLevels;			//--- steering control levels, omega-based but already consider angle levels
    enum orientation_channel { CH_PITCH, CH_YAW, CH_ROLL } OrientationChannel;	//---

	void ManualTryToggleMoonTracking();
	void CalcCurrentOrientLevels();
	void CalcCurrentOmegaLevels();
	void CalcOneCurrentOmegaLevel(orientation_channel Channel);
	void ResetOrientThrusters();
	void SetOrientThrusters();
	void SetChannelThrustersLevel(THGROUP_TYPE THG_POSITIVE, THGROUP_TYPE THG_NEGATIVE, double OmegaLevel);

	// photo
	int PhotoCount;					//+++
	int CurrentPhoto;				//+++

	// photo taking 
	enum photo_step { PH_NONE, PH_WAIT_1, PH_ZOOM, PH_WAIT_2, PH_DEVELOP, PH_MARK, PH_WAIT_3, // 
					  PH_TRANSMIT } 
		PhotoStep;					//+++
	double PhotoTimer;				//+++	separate timer for photo events delays, independent from more generic timer
	double Aperture;				//+++
	double SavedAperture;			//+++
	int ScanLine;					//+++

	SURFHANDLE surfTempPhoto;		//+--
	SURFHANDLE surfPhotos[MAX_PHOTO_COUNT];		//+--
	SURFHANDLE surfBtnL;			//+--
	SURFHANDLE surfBtnR;			//+--
	SURFHANDLE surfBtnTr;			//+--
	SURFHANDLE surfSignal;			//+--
	HDC hDCphoto;					//---

	int SavedHUD;					//+--
	int SavedMFDMode0;				//---
	int SavedMFDMode1;				//---

	void ManualTryTakePhoto();
	void StartPhoto();
	void StepPhoto();

	// transmission
	bool bTransmittedPhotos[MAX_PHOTO_COUNT];	//+++
	bool bSuccessfulPhotos[MAX_PHOTO_COUNT];	//+++
	bool bTransmitEnabled;			//---
	int PercentLost;				//-++
	int ColorRange;					//-++
	int ColorRangeStart;			//-++
	int AutoTransmitPhoto;			//+++		next photo to be transmitted automatically

	// signal strength
	double AltLevel;				//---
	double ElevLevel;				//---
	double SignalLevel;				//---
	double ForcedSignalLevel;		//+++ debug-only

	void ManualTryStartTransmission();
	void StartTransmission();
	void StepTransmission();
	void UpdateSignalLevel();
	void ReleaseSurfaces();
	void StartTransmitting();

	// all animations
	double timer;								//+++   antenna opening delay
	double SimDt;								//---   delta time step, for animation
	bool bAnimate;								//+--
	double tempAnimationPercent;				//+--
    enum animation_type { AT_NONE, AT_ANTENNA, AT_TAPE_EXT, AT_TAPE_ROT, AT_COVER_ROT } CurrentAnimation;
	void DefineAnimations();
	void Animate();

	// antenna animation
	double AntennaRotationPercent;					
	UINT AntennaRotation;						
	void AnimateAntennaRotation();
	void AntennaRotationStart();
	void AntennaRotationEnd();

	// tape extension animation
	double TapeExtensionPercent;					
	UINT TapeExtension;						
	void AnimateTapeExtension();
	void TapeExtensionEnd();

	// tape rotation animation
	double TapeRotationPercent;					
	UINT TapeRotation;						
	void AnimateTapeRotation();
	void TapeRotationEnd();

	// covers rotation animation
	double CoverRotationPercent;					
	UINT CoverRotation;						
	void AnimateCoverRotation();
	void CoverRotationStart();
	void CoverRotationEnd();
	void ManualTryCoverRotationStart();

	// sound
	int SoundLibID;					//---

	// helper functions
	void LocalRot(const VECTOR3 vecGlob, VECTOR3 &vecLoc);
	// rad, rad, rad, rad. rad
	void DistanceFromCoordPairRad(double Lat1, double Lon1, double Lat2, double Lon2, double* pDistance);

	// output
	void DebugSomething();
};





// saving bitmap, Microsotf public code

/******************************Public*Routine******************************\
* SaveBitmapFile
*
*
* Effects: Save pInfo->hBmpSaved into disk specified by pszFileName
*
* Warnings: assumes hBmpSaved is not selected into window's DC other than
*           pInfo->hwnd's DC
*
\**************************************************************************/
BOOL SaveBitmapFile( HDC hDC, HBITMAP hBmp, const char *pszFileName )
{
	int     hFile;
	OFSTRUCT ofReOpenBuff;
	HBITMAP hTmpBmp,
	        hBmpOld;
	int     success;
	BITMAPFILEHEADER bfh;
	PBITMAPINFO pbmi;
	PBYTE   pBits;
	BITMAPINFO bmi;
	PBYTE   pjTmp,
	        pjTmpBmi;
	ULONG   sizBMI;

	success = 0;

	//
	// Let the graphics engine to retrieve the dimension of the bitmap for us
	// GetDIBits uses the size to determine if it's BITMAPCOREINFO or BITMAPINFO
	// if BitCount != 0, color table will be retrieved
	//
	bmi.bmiHeader.biSize = 0x28;	// GDI need this to work
	bmi.bmiHeader.biBitCount = 0;	// don't get the color table
	if ( ( GetDIBits( hDC, hBmp, 0, 0, ( LPSTR ) NULL, &bmi, DIB_RGB_COLORS ) ) == 0 )
	{
		return -1;
	}

	//
	// Now that we know the size of the image, alloc enough memory to retrieve
	// the actual bits
	//
	if ( ( pBits =
		   ( PBYTE ) GlobalAlloc( GMEM_FIXED | GMEM_ZEROINIT,
								  bmi.bmiHeader.biSizeImage ) ) == NULL )
	{
		return -2;
	}

	//
	// Note: 24 bits per pixel has no color table.  So, we don't have to
	// allocate memory for retrieving that.  Otherwise, we do.
	//
	pbmi = &bmi;				// assume no color table
	switch ( bmi.bmiHeader.biBitCount )
	{
		case 24:				// has color table
			sizBMI = sizeof( BITMAPINFOHEADER );
			break;
		case 16:
		case 32:
			sizBMI = sizeof( BITMAPINFOHEADER ) + sizeof( DWORD ) * 3;
			break;
		default:
			sizBMI = sizeof( BITMAPINFOHEADER ) +
					sizeof( RGBQUAD ) * ( 1 << bmi.bmiHeader.biBitCount );
			break;
	}

	//
	// Allocate memory for color table if it is not 24bpp...
	//
	if ( sizBMI != sizeof( BITMAPINFOHEADER ) )
	{
		ULONG   sizTmp;

		//
		// I need more memory for the color table
		//
		if ( ( pbmi =
			   ( PBITMAPINFO ) GlobalAlloc( GMEM_FIXED | GMEM_ZEROINIT,
											sizBMI ) ) == NULL )
		{
			success = -10;
			goto ErrExit1;
		}
		//
		// Now that we've a bigger chunk of memory, let's copy the Bitmap
		// info header data over
		//
		pjTmp = ( PBYTE ) pbmi;
		pjTmpBmi = ( PBYTE ) & bmi;
		sizTmp = sizeof( BITMAPINFOHEADER );
		while ( sizTmp-- )
		{
			*( ( pjTmp )++ ) = *( ( pjTmpBmi )++ );
		}
	}

	//
	// Let's open the file and get ready for writing
	//
	if ( ( hFile = OpenFile( pszFileName, ( LPOFSTRUCT ) & ofReOpenBuff,
							 OF_CREATE | OF_WRITE ) ) == -1 )
	{
		goto ErrExit2;
	}

	//
	// But first, fill in the info for the BitmapFileHeader
	//
	bfh.bfType = 0x4D42;		// 'BM'
	bfh.bfSize =
			sizeof( BITMAPFILEHEADER ) + sizeof( BITMAPINFOHEADER ) + sizBMI +
			pbmi->bmiHeader.biSizeImage;
	bfh.bfReserved1 = bfh.bfReserved2 = 0;
	bfh.bfOffBits = sizeof( BITMAPFILEHEADER ) + sizBMI;

	//
	// Write out the file header now
	//
	if ( _lwrite( hFile, ( LPSTR ) & bfh, sizeof( BITMAPFILEHEADER ) ) == -1 )
	{
		success = -11;
		goto ErrExit3;
	}

	//
	// Bitmap can't be selected into a DC when calling GetDIBits
	// Assume that the hDC is the DC where the bitmap would have been selected
	// if indeed it has been selected
	//
	if ( hTmpBmp =
		 CreateCompatibleBitmap( hDC, pbmi->bmiHeader.biWidth,
								 pbmi->bmiHeader.biHeight ) )
	{
		hBmpOld = (HBITMAP) SelectObject( hDC, hTmpBmp );
		if ( ( GetDIBits
			   ( hDC, hBmp, 0, pbmi->bmiHeader.biHeight, ( LPSTR ) pBits, pbmi,
				 DIB_RGB_COLORS ) ) == 0 )
		{
			success = -12;
			goto ErrExit4;
		}
	}
	else
	{
		success = -13;
		goto ErrExit3;
	}

	//
	// Now write out the BitmapInfoHeader and color table, if any
	//
	if ( _lwrite( hFile, ( LPSTR ) pbmi, sizBMI ) == -1 )
	{
		success = -14;
		goto ErrExit4;
	}

	//
	// write the bits also
	//
	if ( _lwrite( hFile, ( LPSTR ) pBits, pbmi->bmiHeader.biSizeImage ) == -1 )
	{
		success = -15;
		goto ErrExit4;
	}

ErrExit4:
	SelectObject( hDC, hBmpOld );
	DeleteObject( hTmpBmp );
ErrExit3:
	_lclose( hFile );
ErrExit2:
	GlobalFree( pbmi );
ErrExit1:
	GlobalFree( pBits );

	return success;
}


int	signPM(double x)
{
	if (x < 0)
		return -1;
	else
		return 1;
}


