#define		ORBITER_MODULE
#include	"math.h"
#include	"orbitersdk.h"
#include	"stdio.h"
#include	"../math/mymath.h"
#include	"../quaternion/vectors.h"
#include	"../quaternion/matrix.h"
#include	"../quaternion/quaternion.h"
#include	"BlockE.h"


// constructor
R7_third_stage::R7_third_stage(OBJHANDLE hObj, int fmodel): VESSEL2M(hObj, fmodel)
{
	// flags
	bMoonNotCalculated = true;
	bMoonFlyby = false;
	bSwitchFocus = true;
	bManualControl = false;
	bLockPitchOrYaw = true;

	// start with down-looking camera
	CameraDir = CAMERA_DOWN;
	bMoveCamera = false;

	// initialize sodium comet
	CometState = COMET_NONE;
	CometTimer = 0;
	CometStreamLevel = 0;

	SURFHANDLE CometTex = oapiRegisterExhaustTexture("r7_VL\\Contrail_comet"); 

	PARTICLESTREAMSPEC CometStreamSpec = { 0,			// flags, reserved
		0.1,			// srcsize, particle size at creation, m
		0.1,			// srcrate, average particle generation rate, Hz   
		100,			// V0, average particle emission velocity, m/s
		-0.05,			// srcspread, emission velocity distribution randomisation
		10,			// lifetime, average particle lifetime [s]
		10,			// growthrate, particle growth rate [m/s]
		0.1,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::EMISSIVE, PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLOG, 0, 0,
		CometTex};
	AddParticleStream (&CometStreamSpec, _V(0.3, 0.3, 1.4), _V(0,0,1), &CometStreamLevel);

	// 2-nd stage interface
	hR7 = NULL;

	// payload
	PayloadName[0] = 0;
	PayloadMass = 0;

	// fairing
	bFairingNotInitialized = true;
	bFairingAttached = false;
	FairingMass = 0;
	hForFairing = CreateAttachment(TOCHILD, FLSPOINT, FLSDIR, FLSROT, "ForFairing", NOTLOOSE);

	// for autopilot, initialize with big number to prevent premature MECO. 
	CurrentMoonCrossTime = 1000000000;

	// default inclination, deg, will switch to RAD downstream
	Inc = 65.6;

	// MET
	Met = 0;

	// we'll be referencing Earth a lot...
	CalcGeneralPlanetParams();

	// meshes
	int Mesh1 = AddMesh("r7_VL\\BlockE");
	SetMeshVisibilityMode (Mesh1, MESHVIS_ALWAYS);
}

void R7_third_stage::clbkSaveState(FILEHANDLE scn)
{
	// do inherited VESSEL2M stuff before writing custom stuff
	VESSEL2M::clbkSaveState (scn);
	oapiWriteScenario_string (scn, "=========== BLOCK E vars", "");

	// sodium comet: if we have just _any_ comet state, we should save it, only NONE is voided
	if (CometState != COMET_NONE)
		oapiWriteScenario_int (scn, "COMET", CometState);
	if (CometStreamLevel > 0)
		oapiWriteScenario_float (scn, "COMET_LEVEL", CometStreamLevel);
	if (CometTimer > 0)
		oapiWriteScenario_float (scn, "COMET_TIMER", CometTimer);

	// camera view
	if (CameraDir != CAMERA_DOWN)
		oapiWriteScenario_int (scn, "CAMERA", CameraDir);


	// if there is no payload, we are already done: don't write lots of stuff after this point!
	if (PayloadName[0] == 0)
		return;

	// ========= The rest gets written only when we still have payload...

	// MET
	if (Met > 0)
		oapiWriteScenario_float(scn, "MET", Met);

	// manual control override?
	if (bManualControl)
		oapiWriteScenario_int(scn, "MANUAL", bManualControl);

	// keep focus after payload separation?
	if (!bSwitchFocus)
		oapiWriteScenario_int (scn, "SWITCH_FOCUS", bSwitchFocus);

	// don't try to hit the moon?
	if (bMoonFlyby)
		oapiWriteScenario_int(scn, "FLYBY", bMoonFlyby);

	// payload
	oapiWriteScenario_string(scn, "PAYLOAD", PayloadName);
	oapiWriteScenario_vec(scn, "APOINT", PayREF);
	oapiWriteScenario_vec(scn, "PDIR", PayDIR);
	oapiWriteScenario_vec(scn, "PROT", PayROT);

	// target orbit inclination
	if (Inc != 65.6*RAD)
		oapiWriteScenario_float(scn, "INC", Inc*DEG);

	// moon trajectory, if calculated 
	if ( !bMoonNotCalculated)

	{
		sprintf(TrajectoryScn, "%d %.13f %.13f %.13f %.13f %.13f %.13f %.13f %.13f", 
												 bMoonElliptic, 
												 BurnTime, BurnTimeElapsed, BurnTimeRemaining, MoonTravelTime,
												 MyCrossTrueAnomaly, PeriArgumentStart, TargetSemiMajorAxis, MoonCrossR);
		oapiWriteScenario_string(scn, "TRAJECTORY", TrajectoryScn);

		sprintf(TrajectoryScn, "%.13f %.13f %.13f %.13f %.13f %.13f %.13f %.13f %.13f", 
 												 TmEcliToMoon.m11,  TmEcliToMoon.m12, TmEcliToMoon.m13, 
 												 TmEcliToMoon.m21,  TmEcliToMoon.m22, TmEcliToMoon.m23, 
 												 TmEcliToMoon.m31,  TmEcliToMoon.m32, TmEcliToMoon.m33); 
		oapiWriteScenario_string(scn, "ECLI_MOON_MATRIX", TrajectoryScn);
	}
}


void R7_third_stage::ParseScenarioLineEx(char *line, void *status)
{
	if (!strnicmp (line, "MANUAL", 6))
		sscanf (line+6, "%i", &bManualControl);

	else if (!strnicmp (line, "MET", 3))
		sscanf (line+3, "%lf", &Met);
	
	// payload

	else if (!strnicmp (line, "PAYLOAD", 7))
		sscanf (line+7, "%s", &PayloadName);
	else if (!strnicmp (line, "APOINT", 6))
		sscanf (line+6, "%lf %lf %lf", &PayREF.x, &PayREF.y, &PayREF.z);
	else if (!strnicmp (line, "PDIR", 4))
		sscanf (line+4, "%lf %lf %lf", &PayDIR.x, &PayDIR.y, &PayDIR.z);
	else if (!strnicmp (line, "PROT", 4))
		sscanf (line+4, "%lf %lf %lf", &PayROT.x, &PayROT.y, &PayROT.z);

	else if (!strnicmp (line, "INC", 3))
		sscanf (line+3, "%lf", &Inc);

	else if (!strnicmp (line, "FLYBY", 5))
		sscanf (line+5, "%d", &bMoonFlyby);

	// comet (NOTE: "COMET" should be scanned LAST, after other COMET_ values - or it won't give them a chance!

	else if (!strnicmp (line, "COMET_LEVEL", 11))
		sscanf (line+11, "%lf", &CometStreamLevel);
	else if (!strnicmp (line, "COMET_TIMER", 11))
		sscanf (line+11, "%lf", &CometTimer);
	else if (!strnicmp (line, "COMET", 5))
		sscanf (line+5, "%d", &CometState);

	else if (!strnicmp (line, "CAMERA", 6))
		sscanf (line+6, "%d", &CameraDir);

	else if (!strnicmp (line, "SWITCH_FOCUS", 12))
		sscanf (line+12, "%d", &bSwitchFocus);

	// the following will be present in scn file if moon trajectory has been calculated; this is a one-time event
	else if (!strnicmp (line, "TRAJECTORY", 10))
	{
		bMoonNotCalculated = false;
		sscanf (line+10, "%d %lf %lf %lf %lf %lf %lf %lf %lf", 
 			                                     &bMoonElliptic,
			                                     &BurnTime, &BurnTimeElapsed, &BurnTimeRemaining, &MoonTravelTime,
												 &MyCrossTrueAnomaly, &PeriArgumentStart, &TargetSemiMajorAxis, &MoonCrossR);
	}
	else if (!strnicmp (line, "ECLI_MOON_MATRIX", 16))
		sscanf (line+16, "%lf %lf %lf %lf %lf %lf %lf %lf %lf", 
     			                                 &TmEcliToMoon.m11,  &TmEcliToMoon.m12, &TmEcliToMoon.m13, 
     			                                 &TmEcliToMoon.m21,  &TmEcliToMoon.m22, &TmEcliToMoon.m23, 
     			                                 &TmEcliToMoon.m31,  &TmEcliToMoon.m32, &TmEcliToMoon.m33); 

	// VESSEL2M and VESSEL2 stuff
	else 
		VESSEL2M::ParseScenarioLineEx(line, status);
}


// called after all vessel in scn are created, but before first timestep
void R7_third_stage::FinishConfiguration()
{
	// convert inclination to rad
	Inc *= RAD;

	// finish payload attachment, update mass
	PayloadAttach();

	// no nees to save this in the scn file, if the choice is only one of two
	if (bMoonElliptic)
		ArivalDelay = ARRIVAL_DELAY_ELLIPTIC;
	else
		ArivalDelay = ARRIVAL_DELAY_PARABOLIC;

	//VESSEL2M stuff
	VESSEL2M::FinishConfiguration();
}

// this, of course, happens during scenario reading, not in-flight :-)
void R7_third_stage::PayloadAttach()
{
	// create attachment point on us - always, or it screws many things
	hForPayload = CreateAttachment(TOCHILD, PAYPOINT, PAYDIR, PAYROT, "ForPayload", NOTLOOSE);

	// do we have a payload at all?
	if (PayloadName[0] == 0)
		return;

	OBJHANDLE hPayload = oapiGetObjectByName(PayloadName);
	if (!hPayload)
		return;

	// get payload vessel interface, get its mass
	VESSEL	*PayloadIntf;
	PayloadIntf = oapiGetVesselInterface(hPayload);
	PayloadMass = PayloadIntf->GetMass();

	// create attachment point on payload
	OBJHANDLE hToMe = PayloadIntf->CreateAttachment( TOPARENT, PayREF, PayDIR, PayROT, "ToBlockE", NOTLOOSE);

	// Attach!
	AttachChild(hPayload, hForPayload, hToMe);
	SetEmptyMass(EMPTY_MASS + PayloadMass + FairingMass);
}


void R7_third_stage::InitializeFairing()
{
	// flip the fairing initialization flag, so we won't be trying again.
	bFairingNotInitialized = true;

	// do we have fairing?
	OBJHANDLE hFairing = GetAttachmentStatus(hForFairing);
	if (hFairing == NULL)
		return;

	// mark fairing as attached
	bFairingAttached = true;

	// get fairing interface
	hFairingI = oapiGetVesselInterface(hFairing);
	
	// get fairing mass
	FairingMass = hFairingI->GetMass();

	// update full empty mass
	SetEmptyMass(EMPTY_MASS + PayloadMass + FairingMass);
}


// This can be called regardless of whether the fairing is attached or not.
void R7_third_stage::FairingSeparate()
{
	// flip the fairing flag, so we won't be trying again.
	bFairingAttached = false;

	// detach fairing object
	DetachChild(hForFairing, 1);

	// fire fairing separation rockets
	hFairingI->SetThrusterLevel(hFairingI->GetThrusterHandleByIndex(0), 1.0);
	hFairingI->SetThrusterLevel(hFairingI->GetThrusterHandleByIndex(1), 1.0);
	hFairingI->SetThrusterLevel(hFairingI->GetThrusterHandleByIndex(2), 1.0);
	hFairingI->SetThrusterLevel(hFairingI->GetThrusterHandleByIndex(3), 1.0);

	// zero fairing mass
	FairingMass = 0;

	// update full empty mass
	SetEmptyMass(EMPTY_MASS + PayloadMass + FairingMass);
}


void R7_third_stage::clbkSetClassCaps(FILEHANDLE cfg)
{
	// load exhaust textures
	SURFHANDLE etex = oapiRegisterExhaustTexture("r7_VL\\exhaust_yellow"); 
	SURFHANDLE etex_white = oapiRegisterExhaustTexture("r7_VL\\exhaust_crcs.dds"); 

	// set exhaust particle streams %%% can be converted to same style as comet
	PARTICLESTREAMSPEC	psc1;
	PARTICLESTREAMSPEC	psc2;

	memset(&psc1, 0, sizeof(psc1));
	psc1.srcsize = 0.1;
	psc1.srcrate = 20.5;
	psc1.v0 = 15;
	psc1.srcspread = -0.05;
	psc1.lifetime = 0.05;
	psc1.growthrate = 3;
	psc1.atmslowdown = 0.5;
	psc1.ltype = PARTICLESTREAMSPEC::EMISSIVE;
	psc1.levelmap = PARTICLESTREAMSPEC::LVL_PSQRT;
	psc1.lmin = 0;
	psc1.lmax = 0.05;
	psc1.atmsmap = PARTICLESTREAMSPEC::ATM_PLIN;
	psc1.amin = 0.0;
	psc1.amax = 0.0;

	memset(&psc2, 0, sizeof(psc2));
	psc2.srcsize = 0.05;
	psc2.srcrate = 20;
	psc2.v0 = 2;
	psc2.srcspread = -0.05;
	psc2.lifetime = 0.1;
	psc2.growthrate = 2;
	psc2.atmslowdown = 0.5;
	psc2.ltype = PARTICLESTREAMSPEC::EMISSIVE;
	psc2.levelmap = PARTICLESTREAMSPEC::LVL_PSQRT;
	psc2.lmin = 0;
	psc2.lmax = 0.05;
	psc2.atmsmap = PARTICLESTREAMSPEC::ATM_PLIN;
	psc2.amin = 0.0;
	psc2.amax = 0.0;

	// Set physics 
	SetSize(1.0);
	SetSurfaceFrictionCoeff(1e5, 1e5);
	SetLiftCoeffFunc (0);
	SetPMI(PMI_LUNNIK);
	
	// set propellant
	ph_main = CreatePropellantResource(TANK_CAPACITY);

	// set main engine
	th_main[0] = CreateThruster(MAIN_REF, MAIN_DIR, MAIN_THRUST, ph_main, MAIN_ISP);	
	AddExhaust(th_main[0], EX_LENGTH, EX_WIDTH, etex);

	PARTICLESTREAMSPEC MainExhaustStream = { 0,			// flags, reserved
		0.27,			// srcsize, particle size at creation, m
		0.5,			// srcrate, average particle generation rate, Hz   
		50,			// V0, average particle emission velocity, m/s
		-0.05,			// srcspread, emission velocity distribution randomisation
		0.05,			// lifetime, average particle lifetime [s]
		10,			// growthrate, particle growth rate [m/s]
		0.1,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::EMISSIVE, PARTICLESTREAMSPEC::LVL_PSQRT, 0, 0.5, PARTICLESTREAMSPEC::ATM_PLOG, 0, 0 };
	AddExhaustStream(th_main[0], _V(0,0,-1.0), &MainExhaustStream);

	// more consts
	double dZ = 0.1;
	double dS = 0.12;
	double dZ2 = 0.02;

	// set pitch thrusters group
	th_pitch[0] = CreateThruster(CE01_REF, CE01_DIR, CE_THRUST, ph_main, CE_ISP);
	th_pitch[1] = CreateThruster(CE03_REF, CE03_DIR, CE_THRUST, ph_main, CE_ISP);

	AddExhaust(th_pitch[0], EX_LENGTH_BIG, EX_WIDTH_BIG, etex_white);
	AddExhaust(th_pitch[1], EX_LENGTH_BIG, EX_WIDTH_BIG, etex_white);

	AddExhaustStream(th_pitch[0], _V(0,  1.468-dZ*sin(30*RAD), -0.932-dZ*cos(30*RAD)), &psc1);
	AddExhaustStream(th_pitch[1], _V(0, -1.468+dZ*sin(30*RAD), -0.932-dZ*cos(30*RAD)), &psc1);

	// set yaw thrusters group
	th_yaw[0] = CreateThruster(CE02_REF, CE02_DIR, CE_THRUST, ph_main, CE_ISP);
	th_yaw[1] = CreateThruster(CE04_REF, CE04_DIR, CE_THRUST, ph_main, CE_ISP);

	AddExhaust(th_yaw[0], EX_LENGTH_BIG, EX_WIDTH_BIG, etex_white);
	AddExhaust(th_yaw[1], EX_LENGTH_BIG, EX_WIDTH_BIG, etex_white);

	AddExhaustStream(th_yaw[0], _V( 1.468-dZ*sin(30*RAD), 0, -0.932-dZ*cos(30*RAD)), &psc1);
	AddExhaustStream(th_yaw[1], _V(-1.468+dZ*sin(30*RAD), 0, -0.932-dZ*cos(30*RAD)), &psc1);

	// set bank thrusters group
	th_bank_left[0] = CreateThruster(CE05_REF, CE05_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_left[1] = CreateThruster(CE06_REF, CE06_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_left[2] = CreateThruster(CE07_REF, CE07_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_left[3] = CreateThruster(CE08_REF, CE08_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_right[0] = CreateThruster(CE09_REF, CE09_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_right[1] = CreateThruster(CE10_REF, CE10_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_right[2] = CreateThruster(CE11_REF, CE11_DIR, CE_THRUST/2, ph_main, CE_ISP);
	th_bank_right[3] = CreateThruster(CE12_REF, CE12_DIR, CE_THRUST/2, ph_main, CE_ISP);

	AddExhaustStream(th_bank_left[0], _V( 0.042+dS, 1.3, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_left[1], _V( 1.3, -0.042-dS, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_left[2], _V(-0.042-dS, -1.3, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_left[3], _V(-1.3, 0.042+dS, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_right[0], _V(-0.042-dS, 1.3, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_right[1], _V(1.3, 0.042+dS, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_right[2], _V(0.042+dS, -1.3, -0.835+dZ2), &psc2);
	AddExhaustStream(th_bank_right[3], _V(-1.3, -0.042-dS, -0.835+dZ2), &psc2);

	// assemble ALL nozzles into one Main group, so they'll ignite all at once and work together.
	th_main_All[0] = th_main[0];
	th_main_All[1] = th_pitch[0];
	th_main_All[2] = th_pitch[1];
	th_main_All[3] = th_yaw[0];
	th_main_All[4] = th_yaw[1];
	th_main_All[5] = th_bank_left[0];
	th_main_All[6] = th_bank_left[1];
	th_main_All[7] = th_bank_left[2];
	th_main_All[8] = th_bank_left[3];
	th_main_All[9] = th_bank_right[0];
	th_main_All[10] = th_bank_right[1];
	th_main_All[11] = th_bank_right[2];
	th_main_All[12] = th_bank_right[3];
	thg_main = CreateThrusterGroup(th_main_All, 13, THGROUP_MAIN);
}


		// SOUND
void R7_third_stage::clbkPostCreation (void)
{
	SoundLibID = ConnectToOrbiterSoundDLL3(GetHandle());

	SoundOptionOnOff3(SoundLibID, PLAYMAINTHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYATTITUDETHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYLANDINGANDGROUNDSOUND, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAIRSPEED, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAMBIANCEWHENLANDED, true);

	SoundOptionOnOff3(SoundLibID, PLAYCOUNTDOWNWHENTAKEOFF, false);
	SoundOptionOnOff3(SoundLibID, PLAYWHENATTITUDEMODECHANGE, false);
	SoundOptionOnOff3(SoundLibID, PLAYGPWS, false);
	SoundOptionOnOff3(SoundLibID, PLAYHOVERTHRUST, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKINGSOUND, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADARBIP, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKLANDCLEARANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINAIRCONDITIONING, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINRANDOMAMBIANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADIOATC, false);
	SoundOptionOnOff3(SoundLibID, DISPLAYTIMER, false);

	RequestLoadVesselWave3(SoundLibID, SND_PYRO, "Sound\\r7_VL\\Pyro.wav", BOTHVIEW_FADED_FAR);
	RequestLoadVesselWave3(SoundLibID, SND_MOTOR, "Sound\\r7_VL\\Motor.wav", BOTHVIEW_FADED_FAR);
}


void R7_third_stage::clbkPostStep(double simt, double SimDT, double mjd)
{
	SimDt = SimDT;

	// common processing for all VESSEL2M-derived objects
	TimestepForVESSEL2M();

	// camera tracking payload?
	if (bMoveCamera)
		MoveCamera();
	
	// artificial sodium comet, fires when we are way far into space...
	if (CometState > COMET_NONE)
	{
		// time to fire the comet?
		if (CometState == COMET_CHARGED)
		{
			if ( GetAltitude() > 130000000)
				FireComet();
		}

		// time to shut down the comet?
		else if (CometState == COMET_FIRING)
		{
			if ( (CometTimer -= SimDT) < 0)
			{
				CometState = COMET_DONE;
				CometStreamLevel = 0;
				sprintf(oapiDebugString(), "");
			}
			// still running
			else
			{
				// report comet progress
				sprintf(oapiDebugString(), "Sodium comet remaining time: %.0f", CometTimer);

				//block time acceleration
				if ( oapiGetTimeAcceleration() > 1)
					oapiSetTimeAcceleration(1);
			}
		}
	}

	// if we are we airborne, increment the MET timer
	if (Met > 0)
	{
		Met += SimDT;
	}
	else	// if not yet, try to detect the launch contact moment
	{
		VESSELSTATUS vs;
		GetStatus(vs);
		if (vs.status == 0)
			Met += SimDT;
	}

	// one-time event: check if fairing is attached!
	if (bFairingNotInitialized)
		InitializeFairing();

	// Time for fairing separation? this normally happens during 2-nd stage run.
	if ( (bFairingAttached) && (!bManualControl) && ( Met > 161) )
		FairingSeparate();

	// are we still attached to the 2-nd stage? 
	if (bAttached)
	{
		// if we did not grab 2-nd stage interface, grab it now (only once per session)
		if (hR7 == NULL)
		{
			OBJHANDLE hR7Obj = GetAttachmentStatus(GetAttachmentHandle(TOPARENT, 0));
		    VESSEL *hR7I;
			hR7I = oapiGetVesselInterface(hR7Obj);
			hR7 = (R7_second_stage*) hR7I;
		}

		// this indicates our separation from the 2-nd stage
		if (GetAttachmentStatus(GetAttachmentHandle(TOPARENT, 0)) == NULL)
		{
			SetMainLevel(1.0);
			double Ha, Hp, pazimut;
			hR7->GetOrbitParams(Ha, Hp, Inc, pazimut);
			bManualControl = !hR7->GetAutopilot();
			hR7 = NULL;
			bAttached = false;
		}

		return;
	}

	// guidance: don't update if paylaod is already jetisoned.
	if ( PayloadName[0] == 0 )
		return;

	// yet another one-time event: calculate trajectory
	if ( (bMoonNotCalculated) )
	{
		// calculate lunar trajectory
		CalcGeneralMoonParams();
		CalcMoonTouchParams();
		CalcOtherTimes();

		if (bMoonElliptic)
			CalcEllipticPath();
		else
			CalcHyperbolicPath();

		bMoonNotCalculated = false;
	}

	// always update autopilot step: if autopilot is off, we simply don't act upon its data
	ProgramStep(simt, SimDT);

	// manual step
	if (bManualControl)
	{
		ManualControl();

		if (GetNavmodeState(NAVMODE_KILLROT))
			KillRot();
	}
}


int R7_third_stage::clbkConsumeDirectKey(char *keystate)
{
	if (bManualControl)
	{
		// engine shutdown
		if	(
				KEYDOWN (keystate, OAPI_KEY_SUBTRACT) +
				KEYDOWN (keystate, OAPI_KEY_MULTIPLY) 
			)
		{
			SetMainLevel(0.0);
		}

		// engine start
		if	(
				KEYDOWN (keystate, OAPI_KEY_ADD)
			)
		{
			SetMainLevel(1.0);
		}
	}

	else
	{
		// any of the following keys should be disregarded in autopilot mode
		if	(
				KEYDOWN (keystate, OAPI_KEY_ADD) + 
				KEYDOWN (keystate, OAPI_KEY_SUBTRACT) +
				KEYDOWN (keystate, OAPI_KEY_MULTIPLY) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD1) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD2) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD3) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD4) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD6) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD8) +
				KEYDOWN (keystate, OAPI_KEY_NUMPAD5) 
			)
		{
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");
			return 1;
		}
	}

	return 0;
}

int R7_third_stage::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate)
{
	// only process keydown events
	if (!down) 
		return 0; 

	// shift combinations are reserved
	if (KEYMOD_SHIFT (kstate)) 
		return 0; 


	// Control combinations!
	if (KEYMOD_CONTROL (kstate))
	{
		// switch focus to payload after its separation?
		if (key == OAPI_KEY_J) 
		{
			bSwitchFocus = !bSwitchFocus;

			if (bSwitchFocus)
				ShowAnnotationMessage("Switch focus to payload after separation? YES!");
			else
				ShowAnnotationMessage("Switch focus to payload after separation? NO!");
		}

		// manual debug string
		if ( (key == OAPI_KEY_D) )
			DebugSomething();

		return 0; 
	}

	// separate fairing or payload
	if (key == OAPI_KEY_J)
	{
		if (!bManualControl)
			ShowAnnotationMessage("Manual controls disabled. To enable, use M key");

		else if (GetAttachmentStatus(hForFairing))
			FairingSeparate();

		else if (GetAttachmentStatus(hForPayload))
			PayloadSeparate();

		else
			ShowAnnotationMessage("Nothing more to jetison!");
	}

	// toggle internal view 
	if (key == OAPI_KEY_C)
	{
		if (CameraDir == CAMERA_DOWN)
			ToggleOnboardCamera(CAMERA_SIDE);
		else if (CameraDir == CAMERA_SIDE)
			ToggleOnboardCamera(CAMERA_UP);
		else if (CameraDir == CAMERA_UP)
			ToggleOnboardCamera(CAMERA_DOWN);
	}

	// toggle between autopilot and manual control 
	if (key == OAPI_KEY_M) 
		ToggleManualAndAutoplilot();

	return 0;
}


void R7_third_stage::SetPitchMoment(double level)
{
	SetThrusterLevel(th_pitch[0], (1 + level)/2);
	SetThrusterLevel(th_pitch[1], (1 - level)/2);	
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::SetYawMoment(double level)
{
	SetThrusterLevel(th_yaw[0], (1 - level)/2);
	SetThrusterLevel(th_yaw[1], (1 + level)/2);	
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::SetBankMoment(double level)
{
	SetThrusterLevel(th_bank_left[0], (1 + level)/2);
	SetThrusterLevel(th_bank_left[1], (1 + level)/2);
	SetThrusterLevel(th_bank_left[2], (1 + level)/2);
	SetThrusterLevel(th_bank_left[3], (1 + level)/2);

	SetThrusterLevel(th_bank_right[0], (1 - level)/2);
	SetThrusterLevel(th_bank_right[1], (1 - level)/2);
	SetThrusterLevel(th_bank_right[2], (1 - level)/2);
	SetThrusterLevel(th_bank_right[3], (1 - level)/2);
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::ManualControl()
{
	// don't do anything if engine is not running
	if (GetThrusterGroupLevel(thg_main) == 0)
	{
		SetMainLevel(0);
		return;
	}

/*	double lp, lpm, lpp;
	double ly, lym, lyp;
	double lb, lbm, lbp;

	lp = lpm = lpp = 0;
	ly = lym = lyp = 0;
	lb = lbm = lbp = 0;
*/
	double lpp = GetManualControlLevel(THGROUP_ATT_PITCHUP, MANCTRL_ATTMODE, MANCTRL_ANYDEVICE);
	double lpm = GetManualControlLevel(THGROUP_ATT_PITCHDOWN, MANCTRL_ATTMODE, MANCTRL_ANYDEVICE);
	double lp = lpm - lpp;

	double lyp = GetManualControlLevel(THGROUP_ATT_YAWLEFT, MANCTRL_ATTMODE, MANCTRL_ANYDEVICE);
	double lym = GetManualControlLevel(THGROUP_ATT_YAWRIGHT, MANCTRL_ATTMODE, MANCTRL_ANYDEVICE);
	double ly = lym - lyp;

	double lbp = GetManualControlLevel(THGROUP_ATT_BANKLEFT, MANCTRL_ATTMODE, MANCTRL_ANYDEVICE);
	double lbm = GetManualControlLevel(THGROUP_ATT_BANKRIGHT, MANCTRL_ATTMODE, MANCTRL_ANYDEVICE);
	double lb = lbp - lbm;

//		if (fabs(ly) < 0.1) 
			SetPitchMoment(lp);

//		if (fabs(lp) < 0.1) 
			SetYawMoment(ly);

//		if ( (fabs(lp) < 0.1) && (fabs(ly) < 0.1) ) 
			SetBankMoment(lb);
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::SetMainLevel(double level)
{
	SetThrusterGroupLevel(thg_main, level);
	SetThrusterLevel(th_pitch[0], level);
	SetThrusterLevel(th_pitch[1], level);
	SetThrusterLevel(th_yaw[0], level);
	SetThrusterLevel(th_yaw[1], level);
	SetThrusterLevel(th_bank_left[0], level);
	SetThrusterLevel(th_bank_left[1], level);
	SetThrusterLevel(th_bank_left[2], level);
	SetThrusterLevel(th_bank_left[3], level);
	SetThrusterLevel(th_bank_right[0], level);
	SetThrusterLevel(th_bank_right[1], level);
	SetThrusterLevel(th_bank_right[2], level);
	SetThrusterLevel(th_bank_right[3], level);
}


//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::SetHyroplatform(double azimut, double alpha)
{
	Azimut = azimut;
	Alpha = alpha;
}


VECTOR3 R7_third_stage::GetLongAxisDir(double azimut, double alpha)
{
	double	theta;
	VECTOR3	ea, PP, PK, PPK, tg;
	quaternion q1, qi1, v;

	theta = oapiGetPlanetObliquity(hEarth);
	ea = _V(0, cos(theta), sin(theta));
	oapiGetGlobalPos(GetHandle(), &PK);
	oapiGetGlobalPos(hEarth, &PP);
	PPK = PK - PP;
	PPK = PPK*(1/length(PPK));
	tg = crossp(PPK, ea);
	tg = tg*(1/length(tg));

	q1 = _quaternion((float)azimut, _vector3(PPK.x, PPK.y, PPK.z));
	qi1 = qinverse(q1);
	v = _quaternion((float)PI, _vector3(tg.x, tg.y, tg.z));
	q1 = (q1*v)*qi1;
	tg = _V(q1.x, q1.y, q1.z);

	q1 = _quaternion((float)(PI/2 - alpha), _vector3(tg.x, tg.y, tg.z));
	qi1 = qinverse(q1);
	v = _quaternion((float)PI, _vector3(PPK.x, PPK.y, PPK.z));
	q1= (q1*v)*qi1;

	return _V(q1.x, q1.y, q1.z);
}


VECTOR3 R7_third_stage::GetOrientErrors()
{
	VECTOR3	PK, ViP, ViPL, dir;
	double	dB, dP, dY;
	double	sin_dP;
	double	sin_dY;

	dir = GetLongAxisDir(Azimut, Alpha);
	dir = dir*(1/length(dir));
	
	oapiGetGlobalPos(GetHandle(), &PK);
	ViP = PK + dir;
	Global2Local(ViP, ViPL);

	dB = 0;

	sin_dP = ViPL.y/length(ViPL);
	dP=asin(sin_dP);

	sin_dY = ViPL.x/length(ViPL);
	dY=asin(sin_dY);

	// flying backwards?
	if (dir.z > 0.1)
	{
		if (dY > 0)
			dY = -PI + dY;
		else
			dY =  PI + dY;
	}

	return _V(dB, dP, dY);
} 

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
double R7_third_stage::u2(double X, double XX, double w0, double K, double dt)
{
	double upr;
	double dXX;
	double eps = 0.001*RAD;
	double T = 2*dt;
	int n = 0;

	double amax = 1;
	
	double Tmax;
	if (oapiGetTimeAcceleration() == 10)
		Tmax = 10;
	else
		Tmax = 100;

	do
	{
		dXX = XX - sqrt(2*K*fabs(X))*sign(X) - w0;

		upr = -sign(dXX);

		if (K != 0)
			T = abs(dXX/K);

		if (T < Tmax*dt) 
		{
			K = K/2;
			amax = amax/2;
			n++;
		}
	}
	while ( (T < Tmax*dt) && (n < 1000) );

	return upr = upr*amax;
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::SetPitch(double dP, double wp, double t, double dt)
{
	double	Ix;
	VECTOR3	pmi;
	double T, L;
	VECTOR3 omega;
	double	upr;	

	GetPMI(pmi);
	L = ABS( CE01_REF.z*cos(60*RAD)*cos(45*RAD) - CE01_REF.y*sin(60*RAD) );
	Ix = GetMass()*pmi.x;
	GetAngularVel(omega);
	T = 0.5*CE_THRUST;

	upr = -u2(dP, omega.x, wp, T*L/Ix, dt);
	SetPitchMoment(upr);	
}

//-----------------------------------------------------------------------------
//
//-----------------------------------------------------------------------------
void R7_third_stage::SetYaw(double dY, double wy, double t, double dt)
{
	double	Iy;
	VECTOR3	pmi;
	double T, L;
	VECTOR3 omega;
	double	upr;	

	GetPMI(pmi);
	L = ABS( CE01_REF.z*cos(60*RAD)*cos(45*RAD) - CE01_REF.y*sin(60*RAD) );
	Iy = GetMass()*pmi.y;
	GetAngularVel(omega);
	T = 2*CE_THRUST;

	upr = u2(dY, -omega.y, wy, T*L/Iy, dt);
	SetYawMoment(upr);	
}


void R7_third_stage::ProgramStep(double t, double dt)
{
	VECTOR3 R, V, er;
	double r, v, vr, psi, vk;

	GetRelativeVel(hEarth, V);
	GetRelativePos(hEarth, R);
	r = length(R);
	v = length(V);
	er = R*(1/r);
	vr = dotp(V, er);
	
	vk = 100000;
	double dPitch = 0;
	BurnTimeElapsed += dt;

	if (bMoonElliptic)
		CalcEllipticStep();
	else
		CalcHyperbolicStep();

	CalcPeriArgument();
	CalcWantedPeriArgument();

	// be more conservative with flight path correction initially, more aggressive later
	double DP;
	if (BurnTimeElapsed < 200)
		DP = 5;
	else if (BurnTimeElapsed < 300)
		DP = 10;
	else
		DP = 20;

	// simple, but works great!
	if (PeriArgument > WantedPeriArgument)
		dPitch = -DP*RAD;
	else
		dPitch = DP*RAD;

	psi = GetPitch() - GetAOA() + dPitch;

	SetHyroplatform(GetAzimutCorrection(), psi);
	
	VECTOR3 err = GetOrientErrors();
	
	// now we collected all needed data for the autopilot time step.
	// Time to act... but only if autopilot is on!
	if (bManualControl)
		return;

	// act on pitch or yaw error alternatively
	if (bLockPitchOrYaw)
		SetPitch(err.y, 0, t, dt);
	else
		SetYaw(err.z, 0, t, dt);	
	bLockPitchOrYaw = !bLockPitchOrYaw;

	// treat bank differently... normally there is none.
	// We'll need to act only if we are returning to autopilot after having some "fun" in manual.
	FixBank();	

	// on a general flight, limit time accel to x10
	double acc = oapiGetTimeAcceleration();
	if ( acc > 10 )
		oapiSetTimeAcceleration(10);

	// approaching MECO? Get ready: time acceleration, engine thrust, rotation...
	if ( (CurrentMoonCrossTime - MoonTravelTime) < ArivalDelay ) 
	{
		// Block any acceleration
		if ( acc > 1 )
			oapiSetTimeAcceleration(1);
	
		// reduce thrust for better accuracy
		SetMainLevel(0.25);

		// slow down any residual or even navigational rotation - for better payload separation conditions
		SlowRot();
	}

	// MECO time? Fuel depletion? Separate payload!
	if ( (CurrentMoonCrossTime < MoonTravelTime) || (GetPropellantMass(ph_main) == 0) )
		PayloadSeparate();
}


double R7_third_stage::GetAzimutCorrection()
{
	double azimut;
	double phi0, lambda0, r0;
	GetEquPos(lambda0, phi0, r0);
	azimut = asin(cos(Inc)/cos(phi0));

	VECTOR3 PSS, VSS;
	GetRelativePos(hEarth, PSS);
	GetRelativeVel(hEarth, VSS);
	VECTOR3 North = _V(0, 1, 0);
	Ec2Eq(hEarth, PSS);
	Ec2Eq(hEarth, VSS);
	VECTOR3 N1 = crossp(PSS, VSS); N1 = N1*(1/length(N1));
	VECTOR3 N2 = crossp(North, PSS); N2 = N2*(1/length(N2));
	double azimutV = asin(length(crossp(N1, N2)));
	if (azimut < 0) azimutV = -azimutV;
	if(azimut > 90*RAD) azimutV = 90*RAD + azimutV; 
    if(azimut < -90*RAD) azimutV = -90*RAD - azimutV;
	double err = azimutV - azimut;
	//debug_err_AZ = err;
		
	double korr = length(VSS)*err/100.0;
	if(fabs(korr) > 5.0*RAD) korr=sign(korr)*5.0*RAD; 
	azimut -= korr;

	return azimut;
}


void R7_third_stage::KillRot()
{
	const double OMEGA_EPS = 1e-2*RAD;
	
	// get current omega vector
	VECTOR3 Omega;
	GetAngularVel(Omega);
	double OmegaLen = length(Omega);

	if (OmegaLen < OMEGA_EPS)
	{
		// leave some residual rotation. 
		DeactivateNavmode(NAVMODE_KILLROT);
	}
	else
	{
		// invert omega vector to its ort
		Omega = Omega*(1 / OmegaLen);

		// fire engines to counteract the rotation in all three axes 
		SetPitchMoment(Omega.x);
		SetYawMoment(Omega.y);
		SetBankMoment(Omega.z);
	}	
}


void R7_third_stage::SlowRot()
{
	const double OMEGA_EPS = 1.5*RAD;
	
	// get current omega vector
	VECTOR3 Omega;
	GetAngularVel(Omega);
	double OmegaLen = length(Omega);

	// invert omega vector to its ort
	VECTOR3 OmegaOrt = Omega*(1 / OmegaLen);

	// set engines to slow down rotations - each individually!
	if (fabs(Omega.x) > OMEGA_EPS)
		SetPitchMoment(OmegaOrt.x);
	if (fabs(Omega.y) > OMEGA_EPS)
		SetYawMoment(OmegaOrt.y);
	if (fabs(Omega.z) > OMEGA_EPS)
		SetBankMoment(OmegaOrt.z);
}


void R7_third_stage::FixBank()
{
	const double OMEGA_EPS_MAX = 5.0*RAD;
	const double BANK_EPS = 0.1*RAD;
	
	// first, fix omega if rotating too fast

	// get current omega vector
	VECTOR3 Omega;
	GetAngularVel(Omega);
	double OmegaLen = length(Omega);
	VECTOR3 OmegaOrt = Omega*(1 / OmegaLen);
	double OmegaZ = Omega.z;
	double fabsOmegaZ = fabs(OmegaZ);

	// get bank
	double Bank = GetBank();
	double fabsBank = fabs(Bank);

	// bank sensitivity
	double BankEps = BANK_EPS;
	if (oapiGetTimeAcceleration() > 1)
		BankEps *= 10;

	// rotating too fast, any direction? counteract!
	if (fabsOmegaZ > OMEGA_EPS_MAX)
		SetBankMoment(OmegaOrt.z);

	// bank too far off? counteract
	else if (fabs(Bank) > BankEps)
		SetBankMoment(-fabsBank/Bank);

	// bank and omega good and low? stop completely!
	else 
		SetBankMoment(OmegaOrt.z);
}




void R7_third_stage::clbkNavMode (int mode, bool active)
{
	// deactivate: allow silently
	if (!active) 
		return;

	// in manual mode, allow only killrot activation 
	if (bManualControl) 
	{
		if (mode != NAVMODE_KILLROT)
		{
			ShowAnnotationMessage("Modes other than KILLROT are not implemented for Block E.");
			DeactivateNavmode(mode);
		}

		return;
	}

	// in autopilot: activate: override by immediately deactivating it, complain on-screen!
	DeactivateNavmode(mode);
	ShowAnnotationMessage("Manual controls disabled. To enable, use M key.");
}


//============================

// call it from constructor, as we'll need some of these parameters early and/or often...
void R7_third_stage::CalcGeneralPlanetParams()
{
	// get/calc planets data
	hEarth = oapiGetGbodyByName("earth");
	hMoon = oapiGetGbodyByName("moon");

	double MEarth = oapiGetMass(hEarth);
	double MMoon = oapiGetMass(hMoon);

	MuEarth = GRAVITYCONSTANT * MEarth;
	MuBoth = GRAVITYCONSTANT * (MEarth + MMoon); // Moon is a big body, not a point!
}

// Radius-vector touching the Moon at rendesvous point;
// Time of Moon travel to the rendesvous point (prelim);
void R7_third_stage::CalcGeneralMoonParams()
{
	// Get angular momentum vector R*V of the target (Moon) orbit rel to Earth
	VECTOR3 VMoonPos, VMoonVel;	  
	oapiGetRelativeVel(hMoon, hEarth, &VMoonVel);
	oapiGetRelativePos(hMoon, hEarth, &VMoonPos);
	double MoonPos = length(VMoonPos);
	double MoonVel = length(VMoonVel);

	VECTOR3 VMoonAngMomentum = crossp(VMoonPos, VMoonVel); 
	double MoonAngMomentum = length(VMoonAngMomentum);

	// Angular momentum is the "pole" of the orbit, so be it our Z axis.
	// Make z unit vector from it.
	VECTOR3 VMoon1z = VMoonAngMomentum/MoonAngMomentum;    

	// Get eccentricity vector, AKA normalized Laplace vector, points to pericenter
	VECTOR3 VE = VEccentricity(MuBoth, VMoonPos, VMoonVel); 
	eMoon = length(VE);

	// Make x unit vector from it
	VECTOR3 VMoon1x = VE / eMoon;

	// Supplement the orts with y unit vector (points as velocity goes in pericenter point)
	VECTOR3 VMoon1y = crossp(VMoon1z, VMoon1x); 
	double Moon1y = length(VMoon1y);
	VMoon1y /= Moon1y;

	// Make tranlslation matrix from ecliptic to moon orbit reference frame
	TmEcliToMoon = MakeTranslationMatrix(VMoon1x, VMoon1y, VMoon1z);

	// get current Moon true anomaly
	MoonTrueAnomaly = TrueAnomaly(VE, VMoonPos, VMoonVel);

	// get extra orbit params (Orbiter does not have GetElements for planets
	double MoonSpecificMechanicalEnergy = (MoonVel * MoonVel / 2) - (MuBoth / MoonPos);
	double a = -MuBoth / (2 * MoonSpecificMechanicalEnergy);
	pMoon = a * (1 - eMoon * eMoon);
}

// Get our pos, vel (vectors and scalars) in the Moon orbit frame,
// and normal vector (full and unit) to the Moon orbit
void R7_third_stage::RecalcPosVelNormalsInMoonOrbitFrame()
{
	// Get our pos and vel, 
	GetRelativePos(hEarth, VPos);
	GetRelativeVel(hEarth, VVel);

	// Convert to the Moon orbit frame
	VPos = mul(TmEcliToMoon, VPos);
	VVel = mul(TmEcliToMoon, VVel);
	Pos = length(VPos);
	Vel = length(VVel);

	// Get our angular momentum (orbit pole) in the new frame
	VECTOR3 VAngMomentum = crossp(VPos, VVel);   
	double AngMomentum = length(VAngMomentum);

	// Get inclination (z projection angle of ang momentum, <0 if direction of movement is reverse)
	double i = acos(VAngMomentum.z / AngMomentum);  

	// Get normal vector (parallel to the Moon orbit, and lying in the plane of our orbit, points to apocenter)
	VECTOR3 V1z = _V(0, 0, 1);
	VNormal = crossp(V1z, VAngMomentum); 

	// make unit vector from normal vector 
	VNormal1 = VNormal / length(VNormal);
}
		
void R7_third_stage::CalcMoonTouchParams()
{
	RecalcPosVelNormalsInMoonOrbitFrame();

	// Calculate moon orbit R at the point of arrival
		// (and, as a side track, travel time of the Moon from its current location to the rendesvous point)

	// get true anomaly of the crossing point
	double MoonCrossTrueAno = acos(VNormal1.x);
	if (VNormal1.y < 0)
		MoonCrossTrueAno = (2*PI)-MoonCrossTrueAno; 

	// get times of true anomalies
	double MoonTimeOfTrueAnomaly = TimeOfTrueAnomaly(MoonTrueAnomaly, eMoon, MOON_PERIOD);
	double CrossTimeOfTrueAnomaly = TimeOfTrueAnomaly(MoonCrossTrueAno, eMoon, MOON_PERIOD);

		
		// this is what we need to know and store for later use!

	if (CrossTimeOfTrueAnomaly > (MOON_PERIOD/2))
		CrossTimeOfTrueAnomaly = CrossTimeOfTrueAnomaly - MOON_PERIOD;  // negative!

	// travel time of the moon to the rendesvous point
	if (MoonTimeOfTrueAnomaly > (MOON_PERIOD/2))
		MoonTimeOfTrueAnomaly = MoonTimeOfTrueAnomaly - MOON_PERIOD;  // negative!

	MoonTravelTime = CrossTimeOfTrueAnomaly - MoonTimeOfTrueAnomaly;

	// radius of Moon orbit on the crossing vector
	MoonCrossR = pMoon / (1 + eMoon * cos(MoonCrossTrueAno));
}


void R7_third_stage::CalcOtherTimes()
{
	// get current orbit
	ELEMENTS OrbEls;
	ORBITPARAM OrbParam;
	GetElements(hEarth, OrbEls, &OrbParam);

	// calc current time since Peri
	MyTrueAnomaly = OrbParam.TrA;
	double MyTimeSincePeri = OrbParam.T - OrbParam.PeT;

	// calc homan flight time 
	double CurrPeri = OrbParam.PeD;
	double HomanMajorAxis = MoonCrossR + CurrPeri;
	double HomanSemiMajorAxis = HomanMajorAxis / 2;
	double HomanSemiMajorAxis3 = HomanSemiMajorAxis * HomanSemiMajorAxis * HomanSemiMajorAxis;
	double HomanPeriod = 2 * PI * sqrt(HomanSemiMajorAxis3 / MuEarth);

	// calc parabolic flight time 
	double Focal = CurrPeri*2;
	double Focal3 = Focal*Focal*Focal;
	double ParabolicTravelTime;

	ParabolicTravelTime = (1./3.) * sqrt(Focal3/MuEarth) * 
		(
			sqrt(MoonCrossR/CurrPeri - 1)*(1 + MoonCrossR/(2*CurrPeri)) 
			-
			sqrt(Pos/CurrPeri - 1)*(1 + Pos/(2*CurrPeri))
		);

	ParabolicTravelTime -= MyTimeSincePeri;

	bMoonElliptic = MoonTravelTime > ParabolicTravelTime;
}


void R7_third_stage::CalcEllipticPath()
{
	// start with Homan
	double dMoonCrossR = MoonCrossR / 1000; // increment by 0.1 % of homan R
	double PastApoR = MoonCrossR;
	double Fi = 0;
	double PastPeriod = 0;
	double NewSemiMajorAxis = 0;
	bool bCycle = true;
	double NewEta = PI/2;
	int count = 0;

	// find better apo
	while (bCycle)
	{
		double NewApoR = PastApoR + dMoonCrossR;

		NewSemiMajorAxis = (NewApoR + Pos) / 2;
		double NewSemiMajorAxis3 = NewSemiMajorAxis * NewSemiMajorAxis * NewSemiMajorAxis;
		double NewPeriod = 2 * PI * sqrt(NewSemiMajorAxis3 / MuEarth);
		double NewE = (NewSemiMajorAxis - Pos) / NewSemiMajorAxis;
		double NewP = Pos * (1 + NewE);

		NewEta = acos((NewP/MoonCrossR-1)/NewE);
		double NewTravelTime = TimeOfTrueAnomaly(NewEta, NewE, NewPeriod) - TimeOfTrueAnomaly(MyTrueAnomaly, NewE, NewPeriod);
		double deltaTime = NewTravelTime - MoonTravelTime;

		bCycle = ((deltaTime > 0) && !(count++ > 10000));

		PastApoR = NewApoR;
		Fi = PI - NewEta;
		PastPeriod = NewPeriod;
	}

	// true anomaly of cross point in our orbit
	MyCrossTrueAnomaly = NewEta;

	// required velocity at current point
	double PrelimMecoVel = sqrt(MuEarth * (2/Pos - 1/NewSemiMajorAxis));
	double deltaV = PrelimMecoVel - Vel;

	// calc delta m
	double M0 = GetMass();
	double mdot = MAIN_ISP / MAIN_THRUST;  // %%% can add roll engines if needed)
	double mratio = exp(deltaV / MAIN_ISP);
	double Mk = M0 / mratio;
	double Mt = M0 - Mk; 

	// calc burn length
	BurnTime = Mt * mdot;
	BurnTimeElapsed = 0;

	// calc initial argument peri 
	CalcPeriArgument();
	PeriArgumentStart = PeriArgument;

	// select one of two values
	ArivalDelay = ARRIVAL_DELAY_ELLIPTIC;
}

void R7_third_stage::CalcHyperbolicPath()
{
	// start with parabolic vel, increment by dVel m/s
	double PastVel = sqrt((2 * MuEarth) / Pos);
	double dVel = 1; 

	// iteration vars 
	bool bCycle = true;
	int count = 0;

	// result vars
	double NewTravelTime;
	double NewA;
	double CrossTeta;

	// supplemental angle to tangential angle to the hyperbola
	double AngleA = ( (PI/2) - GetPitch() )*2;
	double SinAngleA = sin(AngleA);
	double CosAngleA = cos(AngleA);

	// begin
	while (bCycle)
	{
		// increment velocity from parabolic, at the current radius in the same direction
		double NewVel = PastVel + dVel;

		// some hyperbolic geometry...
		NewA = 1 / ( (NewVel*NewVel)/MuEarth - (2/Pos));
		double NewR2 = 2 * NewA + Pos;
		double New2C = sqrt( Pos*Pos + NewR2*NewR2 - 2*Pos*NewR2*CosAngleA ); 
		double NewC = New2C/2;
		double sinNewTeta = SinAngleA * (NewR2 / New2C);
		double NewTeta = asin(sinNewTeta);
		double cosNewTeta = cos(NewTeta);
		double NewBsq = NewC*NewC - NewA*NewA;
		double NewP = NewBsq / NewA;
		double NewE = NewC / NewA;

		// cross point true anomaly
		double cosCrossTeta = ( ( NewP / MoonCrossR ) - 1 ) / NewE;
		CrossTeta = acos(cosCrossTeta);
		double sinCrossTeta = sin(CrossTeta);

		// some reusable constants
		double Ch = sqrt((NewE + 1) / (NewE - 1));
		double Ct = sqrt(NewA * NewA * NewA / MuEarth);

		// time to current point
		double NewCurrH = 2 * atanh(tan(NewTeta/2) / Ch);
		double CurrPosTime = Ct * (NewE * sinh(NewCurrH) - NewCurrH);

		// time to cross point
		double NewCrossH = 2 * atanh(tan(CrossTeta/2) / Ch);
		double CrossPosTime = Ct * (NewE * sinh(NewCrossH) - NewCrossH);

		// full travel time
		NewTravelTime = CrossPosTime - CurrPosTime;

		// next iteration, if any
		count++;
		PastVel = NewVel;
		bCycle = ( (count < 1000) && (NewTravelTime > MoonTravelTime) );
	}

	// true anomaly of cross point in our orbit
	MyCrossTrueAnomaly = CrossTeta;
	TargetSemiMajorAxis = NewA;

	// required velocity at current point
	double PrelimMecoVel = sqrt(MuEarth * (2/Pos - 1/NewA));
	double deltaV = PrelimMecoVel - Vel;

	// calc delta m
	double M0 = GetMass();
	double mdot = MAIN_ISP / MAIN_THRUST;  
	double mratio = exp(deltaV / MAIN_ISP);
	double Mk = M0 / mratio;
	double Mt = M0 - Mk; 

	// calc burn length
	BurnTime = Mt * mdot;
	BurnTimeElapsed = 0;

	// calc initial argument peri 
	CalcPeriArgument();
	PeriArgumentStart = PeriArgument;

	// select one of two values
	ArivalDelay = ARRIVAL_DELAY_ELLIPTIC;
}



void R7_third_stage::CalcEllipticStep()
{
	// get current orbit
	ELEMENTS OrbEls;
	ORBITPARAM OrbParam;
	GetElements(hEarth, OrbEls, &OrbParam);

	// calc current time since Peri
	double CurrentTrueAnomaly = OrbParam.TrA;
	double CurrentTimeSincePeri = OrbParam.T - OrbParam.PeT;

	// calc true anomaly of moon cross radius
	double NewP = OrbParam.PeD * (1 + OrbEls.e);
	double MoonCrossEta = acos((NewP/MoonCrossR-1)/OrbEls.e);
	CurrentMoonCrossTime = TimeOfTrueAnomaly(MoonCrossEta, OrbEls.e, OrbParam.T) - CurrentTimeSincePeri;
}

void R7_third_stage::CalcHyperbolicStep()
{
	// get current orbit
	ELEMENTS OrbEls;
	ORBITPARAM OrbParam;
	GetElements(hEarth, OrbEls, &OrbParam);


	// required velocity at current point
	double PrelimMecoVel = sqrt(MuEarth * (2/Pos - 1/TargetSemiMajorAxis));
	double deltaV = PrelimMecoVel - Vel;

	// calc delta m
	double M0 = GetMass();
	double mdot = MAIN_ISP / MAIN_THRUST;  // %%% can add roll engines if needed)
	double mratio = exp(deltaV / MAIN_ISP);
	double Mk = M0 / mratio;
	double Mt = M0 - Mk; 

	// calc burn length
	BurnTimeRemaining = Mt * mdot;

	// limit time acceleration to x10
	double acc = oapiGetTimeAcceleration();
	if (!bManualControl)
	{
		if (acc > 10)
			oapiSetTimeAcceleration(10);
	}

	// are we hyperbolic already? don't waste time if not yet.
	if (OrbEls.e <= 1)
		return;

	// the folloing runs only when we go over the escape velocity

	// some reusable constants
	double a = fabs(OrbEls.a);
	double Ch = sqrt((OrbEls.e + 1) / (OrbEls.e - 1));
	double Ct = sqrt(a * a * a / MuEarth);

	// calc current time since Peri
	double CurrentTrueAnomaly = OrbParam.TrA;

	double CurrH = 2 * atanh(tan(CurrentTrueAnomaly/2) / Ch);
	double CurrentTimeSincePeri = Ct * (OrbEls.e * sinh(CurrH) - CurrH);

	// calc true anomaly of moon cross radius
	double NewP = a * (OrbEls.e * OrbEls.e - 1);
	double MoonCrossTeta = acos((NewP/MoonCrossR-1)/OrbEls.e);

	double CrossH = 2 * atanh(tan(MoonCrossTeta/2) / Ch);
	CurrentMoonCrossTime = Ct * (OrbEls.e * sinh(CrossH) - CrossH);

	// time
	CurrentMoonCrossTime -= CurrentTimeSincePeri;

	CurrentMoonCrossTime+=(3600/2);
}

void R7_third_stage::CalcPeriArgument()
{
	RecalcPosVelNormalsInMoonOrbitFrame();
			
	// eccentricity vector of our orbit (in the Moon ref frame)
	VECTOR3 VE = VEccentricity(MuEarth, VPos, VVel);

	// unit vector of the eccentricity dir
	VECTOR3 VE1 = VE / length(VE);    

	//angle
	PeriArgument = acos(dotp(VNormal1, VE1));
	if (VE1.z < 0)
		PeriArgument = -(PI-PeriArgument);
	else
		PeriArgument = PI-PeriArgument;

	// Rotate argument: we need arbitrary point on the ellipse, not the apocenter!
	// Also, empirically compensate for a slight drift of the original MyCrossTrueAnomaly due to higher perigee.
	// With the resulting PeriArgument, we should hit the Moon!
	if (bMoonElliptic)
		PeriArgument += ( (PI - MyCrossTrueAnomaly) - 1*RAD );
	else
		PeriArgument += ( (PI - MyCrossTrueAnomaly) - 1.5*RAD );

	// By default, we want to hit the Moon right on.
	// however, for the Luna-1 or Luna-3 trajectory types we don't want to hit the Moon, 
	// but rather try to pass 6.5 Mm over the South Pole.
	// So, we slightly overrotate the flight path here.
	if (bMoonFlyby)
		PeriArgument += (0.65 * RAD);
}


void R7_third_stage::CalcWantedPeriArgument()
{
	if (bMoonElliptic)
		WantedPeriArgument = PeriArgumentStart * (1 - BurnTimeElapsed/BurnTime);
	else
	{
		double PlannedBurnTimeRemaining = BurnTime - BurnTimeElapsed;
		double BurnTimeUpdate = BurnTimeRemaining - PlannedBurnTimeRemaining;
		double BurnTimeUpdated = BurnTime + BurnTimeUpdate;

		WantedPeriArgument = PeriArgumentStart * (1 - BurnTimeElapsed/BurnTimeUpdated);
	}
}




// SetCameraDefaultDirection has no effect when called from clbkSetClassCaps - bug in Orbiter? 
// Enforce camera dir switch here:
bool R7_third_stage::clbkLoadGenericCockpit ()
{
	oapiSetDefRCSDisplay(0);
	ToggleOnboardCamera(CameraDir);
	return true;
}


void R7_third_stage::FireComet()
{
	// don't do anything if comet is already firing or done 
	if (CometState != COMET_CHARGED)
		return;

	CometState = COMET_FIRING;
	CometStreamLevel = 1;
	CometTimer = 60;
	
	ShowAnnotationMessage("Artificial sodium comet fired!");

	if ( oapiGetTimeAcceleration() > 1)
		oapiSetTimeAcceleration(1);
}

void R7_third_stage::ToggleManualAndAutoplilot()
{
	// do not allow toggling during pre-burn!
	if (bAttached)
	{
		ShowAnnotationMessage("Manual/Autopilot controls can't be toggled for Block E until its separation!");
		return;
	}

	// do not allow toggling after burn!
	if (PayloadName[0] == 0)
	{
		ShowAnnotationMessage("Manual/Autopilot controls can't be toggled after mission completion!");
		return;
	}

	// toggle flag
	bManualControl = !bManualControl;

	// implement switching to the appropriate mode
	if (bManualControl)
		SetManualControl();
	else
		SetAutopilotControl();
}

void R7_third_stage::SetManualControl()
{
	ShowAnnotationMessage("Manual controls enabled. To return to autopilot, use M key");
	sprintf(oapiDebugString(), "");
}


void R7_third_stage::SetAutopilotControl()
{
	ShowAnnotationMessage("Manual controls disabled. To enable, use M key");

	// restore main engine to full throttle
	SetMainLevel(1);

	// disable any possible residual navmodes
	DeactivateNavmode(NAVMODE_KILLROT);
	DeactivateNavmode(NAVMODE_HLEVEL);
	DeactivateNavmode(NAVMODE_RETROGRADE);
	DeactivateNavmode(NAVMODE_NORMAL);
	DeactivateNavmode(NAVMODE_ANTINORMAL);
	DeactivateNavmode(NAVMODE_HOLDALT);
	DeactivateNavmode(NAVMODE_PROGRADE);
}


// by J key, by autopilot or by accident :-)
void R7_third_stage::PayloadSeparate()
{
	// do we have someting to separate?
    if (!GetAttachmentStatus(hForPayload))
		return;

	// shutdown main engine if running
	SetMainLevel(0.0);

	//separation velocity
	double VDetach = 0.1; 

	// separate whatever is attached to the payload attacment, indiscriminatory
	DetachChild(hForPayload, VDetach);

	// update empty mass
	PayloadMass = 0;
	SetEmptyMass(EMPTY_MASS + PayloadMass + FairingMass);

	// get payload handle by name before killing the name 
	OBJHANDLE hPayload = oapiGetObjectByName(PayloadName);
	PayloadName[0] = 0;

	// change focus to payload - only if we have focus in the first place!
	if ( (bSwitchFocus) && (oapiGetFocusObject() == GetHandle()) )
	{
		// Make sure that the originally attached object still exists.
		// This won't work if the object was destroyed, and only some piece remains...
		// but it won't CTD either, just keeps the focus at L.
		if (hPayload) 
		{
			VESSEL	*PlIntf;
			PlIntf = oapiGetVesselInterface(hPayload);

			// one last check: is payload allowed to receive focus?
			if (PlIntf->GetEnableFocus()) 
				oapiSetFocusObject(hPayload);
		}
	}

	// indicate autopilot mode to payloat in case it needs it (and can understand it)
	if (bManualControl)
		SendVesselMessage(hPayload, VM_AUTOPILOT_OFF);

	// force-switch ourselves to autopilot (or rather prevent manual modes)
	bManualControl = false;

	// start camera rotation
	if (CameraDir == CAMERA_SIDE)
		StartCameraRotation();

	// pyro sound
	PlayVesselWave3(SoundLibID, SND_PYRO);
}



void R7_third_stage::ToggleOnboardCamera(CameraDirs Dir)
{
	// stop any residual rotation
	bMoveCamera = false;
	StopVesselWave3(SoundLibID, SND_MOTOR);

	// update direction var
	CameraDir = Dir;

	if (Dir == CAMERA_SIDE)
	{
		SetCameraOffset(_V(0.0, 1.4, 1.5));
		SetCameraDefaultDirection(_V(0, -sqrt(1-CAMERA_Z_SIDE*CAMERA_Z_SIDE), CAMERA_Z_SIDE));
	}

	if (Dir == CAMERA_UP)
	{
		SetCameraOffset(_V(0.0, 1.4, 1.5));
		SetCameraDefaultDirection(_V(0,0,1));
	}

	else if (Dir == CAMERA_DOWN)
	{
		SetCameraOffset(_V(1.0, -1.0, 1.0));
		SetCameraDefaultDirection(_V(0,0,-1));
	}
}

void R7_third_stage::StartCameraRotation()
{
	bMoveCamera = true;
	PlayVesselWave3(SoundLibID, SND_MOTOR, LOOP, 180);
	CameraZ = CAMERA_Z_SIDE;
	CameraVector.x = 0;
}

void R7_third_stage::MoveCamera()
{
	// camera travel in timestep, rad
	double dAngle = CAMERA_SPEED_RAD * SimDt;

	// past camera angle
	double CameraAngle = asin(CameraZ);

	// new camera angle
	CameraAngle += dAngle;

	// done?
	if (CameraAngle > (PI/2))
	{
		CameraAngle = (PI/2);
		StopCameraRotation();
	}

	// new position of z-component
	CameraZ = sin(CameraAngle);

	// new dir vector;
	CameraVector.z = CameraZ;
	CameraVector.y = -sqrt(1-CameraZ*CameraZ);

	// point camera
	SetCameraDefaultDirection(CameraVector);
}


void R7_third_stage::StopCameraRotation()
{
	// clear rotation flag
	bMoveCamera = false;

	StopVesselWave3(SoundLibID, SND_MOTOR);

	// update internal camera dir state
	if (CameraDir == CAMERA_SIDE)
		CameraDir = CAMERA_UP; 
}



void R7_third_stage::DebugSomething()
{
	sprintf(oapiDebugString(), "o1 %f o2 %f i1 %d", dOutp1, dOutp2, iOutp1);
}




DLLCLBK VESSEL *ovcInit(OBJHANDLE hvessel, int flightmodel)
{
	return new R7_third_stage(hvessel, flightmodel);
}

DLLCLBK void ovcExit(VESSEL *vessel)
{
	if (vessel) delete (R7_third_stage*) vessel;
}

