#include <stdio.h>
#include <time.h>
#include "vessel2m.h"
#include "..\r7_wreck_VL\wreck.h"


VESSEL2M::VESSEL2M (OBJHANDLE hObj, int fmodel): VESSEL2 (hObj, fmodel)
{
	// set default structural limits, in case derived object forgets to do so...
	VvCrash = 1; 
	VhCrash = 1;
	VMaxBlast = 1000; 
	QMax = 20000;

	// focus defaults
	bFocusOffByDefault = false;
	bFocus = true;

	// attached default
	bAttached = false;
	
	// seed random generator
	srand(int(time(NULL)));

	// generate a small cloud of dust at touchdown
	PARTICLESTREAMSPEC Dust = {
		0,			// flags, reserved
		0.1,			// srcsize, particle size at creation, m
		20,			// srcrate, average particle generation rate, Hz   
		1,			// V0, average particle emission velocity, m/s
		0,			// srcspread, emission velocity distribution randomisation
		4,			// lifetime, average particle lifetime [s]
		3,			// growthrate, particle growth rate [m/s]
		5,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 
		0, 
		0.5,
		PARTICLESTREAMSPEC::ATM_FLAT, 
		1.22, 
		1.25
	};
	bGeneratingDust = false;
	bLandedPermanently = false;
	DustTimer = -1;
	DustLevel = 0;
	AddParticleStream (&Dust, _V(0,0,0), _V(0,1,0), &DustLevel);

	// annotation window
	MessagePort = oapiCreateAnnotation ( true, 0.5, _V(1, 1, 1) );
	oapiAnnotationSetPos (MessagePort, 0, 0.1, 1, 1);
	msgStartTime = 0;
	bMessageDisplayed = false;
}

// destructor
VESSEL2M::~VESSEL2M ()
{
	// clean up output
	oapiAnnotationSetText (MessagePort, "");
	sprintf(oapiDebugString(), "");
}


void VESSEL2M::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
    char *line;

	while (oapiReadScenario_nextline (scn, line)) 
		ParseScenarioLineEx(line, status);

	FinishConfiguration();
}


void VESSEL2M::ParseScenarioLineEx(char *line, void *status)
{
	// focus
	if (!_strnicmp (line, "FOCUS", 5))
		sscanf (line+5, "%d", &bFocus);

	// default Orbiter processing, but intercept STATUS and ATTACHED 
	else 
	{
		// are we landed? we want to know to avoid dusting
		if (!_strnicmp (line, "STATUS", 6))
		{
			if (!_strnicmp (line+7, "Landed", 6))
				bLandedPermanently = true;
		}

		// are we attached to parent?
		else if (!_strnicmp (line, "ATTACHED", 8))
			bAttached = true;

		//pass to Orbiter's generic parser
		VESSEL2::ParseScenarioLineEx (line, status);
	}
}


void VESSEL2M::clbkSaveState (FILEHANDLE scn)
{
	// save default vessel parameters
	SaveDefaultState (scn);

	// focus enabled?
	bFocus = GetEnableFocus();

	// only store focus in scenario if different from default
	if (bFocusOffByDefault)
	{
		if (bFocus) 
			oapiWriteScenario_int (scn, "FOCUS", bFocus);
	}
	else
	{
		if (!bFocus) 
			oapiWriteScenario_int (scn, "FOCUS", bFocus);
	}
}


void VESSEL2M::FinishConfiguration()
{
	SetEnableFocus(bFocus == 1);
}



// this NEVER gets ignored internally!
// To ignore VESSEL2 processing, just don't call it from the derived object.
void VESSEL2M::TimestepForVESSEL2M()
{
	// don't do any processing when still attached
	if (bAttached)
		return;

	// message display timer - always give it a chance to finish
	if (bMessageDisplayed)
	{
		if ( (oapiGetSysTime() - msgStartTime) > MSG_DISPLAY_TIME )
		{
			oapiAnnotationSetText (MessagePort, "");
			bMessageDisplayed = false;
		}
	}

	if (GroundContact())
	{
		// ignore dusting if completely landed; we don't want it while sitting on the ground!
		if (!bLandedPermanently)
			if (DustTimer <= 0) 
				StartDust();

		GetLandingVelocities();

		if (V > VMaxBlast) 
			Explode();

		else if ( (Vv > VvCrash) || (Vh > VhCrash) )
		{
			if (bFuelExplode())
  				Explode();
			else
				Crash();
		}

		// Non-destructive landing. If there is any damage, derived vessel will handle it on her own.
		// Let vessel do it own landing routine
		else 
			Land();
	}

	if (GetDynPressure() > QMax)
		Burn();

	if (bGeneratingDust)
		if ( (DustTimer-=oapiGetSimStep()) < 0 )
			StopDust();
}


void VESSEL2M::GetLandingVelocities()
{
	// read velocity in horizon frame, obtain horisontal and vertical velocity scalars from it
	VECTOR3 HAV; 
	GetHorizonAirspeedVector(HAV);

	V = length(HAV);
	Vv = fabs(HAV.y);
	Vh = sqrt(HAV.x*HAV.x + HAV.z*HAV.z);
}


// returns true if fuel mass is more than 10% of the vessel weight
bool VESSEL2M::bFuelExplode()
{
	double FuelMass = 0;
	for (DWORD i=0; i<GetPropellantCount(); i++)
		FuelMass += GetPropellantMass(GetPropellantHandleByIndex(i));

	return ((FuelMass / GetMass()) > 0.1);
}


void VESSEL2M::Explode()
{
	ReplaceWithWreck(EXPLODE);
}


void VESSEL2M::Crash()
{
	ReplaceWithWreck(NO_EXPLODE);
}


void VESSEL2M::Burn()
{
	// kind of repeat of the logic, but here we are giving chance to run customized vessel methods
	if (bFuelExplode())
		Explode();
	else
		Crash();
}


void VESSEL2M::Land()
{
}


void VESSEL2M::ReplaceWithWreck(bool bExplode, bool bGenericCrater)
{
	// 1. PREPARATION

	// if we have any children, drop them 
	for (DWORD i = 0; i < AttachmentCount(TOCHILD); i++)
	{
		ATTACHMENTHANDLE hAttToChild = GetAttachmentHandle(TOCHILD, i);
		OBJHANDLE hChild = GetAttachmentStatus(hAttToChild);
		if (hChild != NULL)
			DetachChild(hAttToChild, 0);
	}

	// do we have atmosphere around?
	double Pressure;
	double Density;
	oapiGetAtmPressureDensity(GetHandle(), &Pressure, &Density);

	// prepare a name for a wreck object 
	char Name[100];
	strcpy(Name, GetName());
	strcat(Name, "-Wreck");

	// clone wreck status from the current one
	VESSELSTATUS vs;
	GetStatus(vs);

	// force landing by updating vessel status
	if (GroundContact())
	{
		double Longitude, Latitude, Radius;
		GetEquPos(Longitude, Latitude, Radius);

		vs.status = 1;
		vs.vdata[0].x = Longitude;
		vs.vdata[0].y = Latitude;
		vs.vdata[0].z = 0;
	}

	// 2. WRECK OBJECT CREATION

	// create a wreck object - this will run its constructor 
	void *hWreck;
	hWreck = oapiCreateVessel(Name, "r7_VL\\Wreck", vs); 

	// get wreck object interface
	VESSEL *VesselIntf;
	R7_wreck* Intf;
	VesselIntf = oapiGetVesselInterface(hWreck);
	Intf = (R7_wreck*)VesselIntf;

	// 3. WRECK OBJECT INITIALIZATION 
	// this is a runtime alternative to the loading object from scenario

	// blast? smoke? nothing?
	if (bExplode)
		Intf->BlastState = Intf->TOBLAST; 
	else if (Pressure > 0)
		Intf->BlastState = Intf->TOSMOKE;
	else
		Intf->BlastState = Intf->TODUST;

	// update and slightly modify physics based on the originating object (self)
	Intf->SetSize(GetSize());
	Intf->SetEmptyMass(GetEmptyMass() * 0.95); // decrease mass by 5 %
	VECTOR3 cs;
	GetCrossSections(cs);
	Intf->SetCrossSections(cs * 1.1); // increase cross-section by 10 %

	// load crater mesh?
	Intf->bLoadMesh = (Pressure == 0);

	// can focus wreck?
	Intf->bFocus = bFocus;

	// landed flag
	Intf->bFlying = vs.status != 1;

	// 4. SWITCH TO WRECK OBJECT 

	if (oapiGetFocusObject() == GetHandle())
	{
		// cleanup output
  		sprintf(oapiDebugString(), "");

		// switch focus
		oapiSetFocusObject(hWreck);
	}

	// finalize wreck object initialization, like after loading from scenario
	Intf->bJustDeployed = true;


	// 5. CLEANUP OLD OBJECT 

	// Finally, delete myself!
	oapiDeleteVessel (GetHandle(), 0);
}



void VESSEL2M::ShowAnnotationMessage(char *Message)
{
	// don't show message if still attached - we won't be able to remove it until timestep is activated on separation!
	if (bAttached)
		return;

	oapiAnnotationSetText (MessagePort, Message);
	msgStartTime = oapiGetSysTime();
	bMessageDisplayed = true;
}


void VESSEL2M::SendVesselMessage(OBJHANDLE hTarget, VesselMessage VM)
{
	VESSEL *Intf;

	if (hTarget != NULL)
	{
		Intf = oapiGetVesselInterface(hTarget);
		if (Intf != NULL)
			((VESSEL2M*)Intf)->GetVesselMessage(VM);
	}
}


void VESSEL2M::GetVesselMessage(VesselMessage VM)
{
	;
}



double VESSEL2M::DRand(double from, double  span)
{
	return ((double)rand() / (double)RAND_MAX) * span + from;
}

int VESSEL2M::IRand(int from, int  span)
{
	return (int)( (double)rand() / (double)RAND_MAX * (double)span ) + from;
}


void VESSEL2M::StartDust()
{
	bGeneratingDust = true;
	DustTimer = 0.2;
	DustLevel = 1;
}


void VESSEL2M::StopDust()
{
	bGeneratingDust = false;
	DustTimer = -1;
	DustLevel = 0;

	// if by the time of the "dust-off" we are still sitting on the ground, consider us landed.
	bLandedPermanently = GroundContact();
}