#define ORBITER_MODULE


#include "wreck.h"
#include "OrbiterSoundSDK35.h"
#include "stdio.h"
#include "math.h"


// constructor
R7_wreck::R7_wreck (OBJHANDLE hObj, int fmodel): VESSEL2 (hObj, fmodel)
{
	// by default, we assume loading "craters" on the Moon from scenario, no focus, no smoke 
	bJustDeployed = false;
	bFocus = false;  //most wreck fragments don't allow focus
	BlastState = NONE;
	bLoadMesh = true;
	hCraterMesh = NULL;

	// this default is different from others, it is just easier to override it this way
	bFlying = true;
}



// called when reading the scenario file - this does not happen on runtime creation
void R7_wreck::clbkLoadStateEx (FILEHANDLE scn, void *status)
{
	double x, y, z;
	VECTOR3 V;

	//Read the scenario file line by line
    char *line;
	while (oapiReadScenario_nextline (scn, line)) 
	{
		if (!_strnicmp (line, "FOCUS", 5))
			sscanf (line+5, "%d", &bFocus);

		else if (!_strnicmp (line, "LOADMSH", 7))
			sscanf (line+7, "%d", &bLoadMesh);

		else if (!_strnicmp (line, "SMOKE", 5))
		{
			sscanf (line+5, "%d", &BlastState);

			// we have to update static states to starting ones)
			if (BlastState == BLASTING) 
				BlastState = TOBLAST;
			else if (BlastState == SMOKING) 
				BlastState = TOSMOKE;
		}

		// physics, only when flying in atmoshpere

		else if (!_strnicmp (line, "SIZE", 4))
		{
			sscanf (line+4, "%lf", &x);
			SetSize(x);
		}

		else if (!_strnicmp (line, "MASS", 4))
		{
			sscanf (line+4, "%lf", &x);
			SetEmptyMass(x);
		}

		else if (!_strnicmp (line, "CROSSSEC", 8))
		{
			sscanf (line+8, "%lf %lf %lf", &x, &y, &z);
			V.x = x; V.y = y; V.z = z;
			SetCrossSections(V);
		}

		else // default Orbiter parser, but intercept landing status
		{
			// are we landed? 
			if (!_strnicmp (line, "STATUS", 6))
			{
				if (!_strnicmp (line+7, "Landed", 6))
				bFlying = false;
			}
			
			ParseScenarioLineEx (line, status);
		}
	}

	// additional post-load configuration here
	FinishConfiguration();
}



// Save status to scenario file
void R7_wreck::clbkSaveState (FILEHANDLE scn)
{
	// save default vessel parameters
	SaveDefaultState (scn);

	// visual separator
	oapiWriteScenario_string (scn, "===========", "");

	// save custom parameters
	if (bFocus)
		oapiWriteScenario_int (scn, "FOCUS", bFocus);

	if (!bLoadMesh)
		oapiWriteScenario_int (scn, "LOADMSH", bLoadMesh);

	if (BlastState != NONE)
		oapiWriteScenario_int (scn, "SMOKE", BlastState);

	// if still flying, save physics
	if (bFlying)
	{
		oapiWriteScenario_float (scn, "SIZE", GetSize());
		oapiWriteScenario_float (scn, "MASS", GetEmptyMass());
		VECTOR3 cs;
		GetCrossSections(cs);
		oapiWriteScenario_vec (scn, "CROSSSEC", cs);
	}
}


// this function runs both for scenario and runtime creation
void R7_wreck::FinishConfiguration()
{	
	// start with focus
	SetEnableFocus(bFocus == 1);

	// touchdown points
	SetTouchdownPoints (_V( 1, 0, 0), _V( -1, 0, 1), _V(-1, 0, -1));

	// mesh
	if (bLoadMesh)
	{
		// load crater mesh
		VECTOR3 MeshOffset = _V(0, 0, 0);
		int MeshID;

		MeshID = AddMesh ("r7_VL\\WreckCrater", &MeshOffset);

		SetMeshVisibilityMode (MeshID, MESHVIS_ALWAYS);

		HideShadows();
	}
/*
	SoundLibID = ConnectToOrbiterSoundDLL3(GetHandle());

	SoundOptionOnOff3(SoundLibID, PLAYMAINTHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYATTITUDETHRUST, true);
	SoundOptionOnOff3(SoundLibID, PLAYLANDINGANDGROUNDSOUND, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAIRSPEED, true);
	SoundOptionOnOff3(SoundLibID, PLAYWINDAMBIANCEWHENLANDED, true);

	SoundOptionOnOff3(SoundLibID, PLAYCOUNTDOWNWHENTAKEOFF, false);
	SoundOptionOnOff3(SoundLibID, PLAYWHENATTITUDEMODECHANGE, false);
	SoundOptionOnOff3(SoundLibID, PLAYGPWS, false);
	SoundOptionOnOff3(SoundLibID, PLAYHOVERTHRUST, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKINGSOUND, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADARBIP, false);
	SoundOptionOnOff3(SoundLibID, PLAYDOCKLANDCLEARANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINAIRCONDITIONING, false);
	SoundOptionOnOff3(SoundLibID, PLAYCABINRANDOMAMBIANCE, false);
	SoundOptionOnOff3(SoundLibID, PLAYRADIOATC, false);
	SoundOptionOnOff3(SoundLibID, DISPLAYTIMER, false);

	RequestLoadVesselWave3(SoundLibID, SND_CRASH, "Sound\\Vessel\\Crash.wav", BOTHVIEW_FADED_FAR);

	PlayVesselWave3(SoundLibID, SND_CRASH);
*/

	// finally, make sure we run only once
	bJustDeployed = false;
}




// simple, fast, short
void R7_wreck::clbkPostStep(double simt, double simdt, double mjd)
{
	// Now that we know additional details, we can finalize the configuration and physics
	// to differentiate between landed or flying states.
	// This only needs to run once, before the very first time step.
	if (bJustDeployed)
		FinishConfiguration();

	// blast/smoke control
	if ( (BlastState != SMOKING) && (BlastState != NONE) )
	{
		if (BlastState == TOBLAST)
			StartBlast();

		else if ((BlastState == BLASTING) && (StopBlastTime < simt))
			StopBlast();

		else if (BlastState == TOSMOKE)
			StartSmoke();

		else if (BlastState == TODUST)
			StartDust();

		else if ((BlastState == DUSTING) && (StopDustTime < simt))
			StopDust();
	}

	// land and stop
	if ((bFlying) && (GroundContact()))
		Land();
}


void R7_wreck::Land()
{
	// force landing by updating vessel status
	double Longitude, Latitude, Radius;
	VESSELSTATUS vs;

	GetEquPos(Longitude, Latitude, Radius);
	GetStatus(vs);

	vs.status = 1;
	vs.vdata[0].x = Longitude;
	vs.vdata[0].y = Latitude;
	vs.vdata[0].z = 0;

	DefSetState(&vs);

	PlayVesselWave3(SoundLibID, SND_CRASH);

	bFlying = false;
}


void R7_wreck::StartBlast()
{
	// create a very quickly expanding white "fireball"
	PARTICLESTREAMSPEC BlastFire = {
		0, 
		0.1, 
		4, 
		0, 
		0, 
		2, 
		100, 
		0, 
		PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_FLAT, 
		1, 
		1,
		PARTICLESTREAMSPEC::ATM_FLAT, 
		1, 
		1
	};

	AddParticleStream (&BlastFire, _V(0,0,0), _V(0,1,0), &BlastLevel);
	BlastLevel = 1;
	BlastState = BLASTING;

	// change flags to stop the blast after one second
	StopBlastTime = oapiGetSimTime() + 1;
}



void R7_wreck::StopBlast()
{
	BlastLevel = 0;

	double Pressure;
	double Density;
	oapiGetAtmPressureDensity(GetHandle(), &Pressure, &Density);

	// do we have atmosphere around? 
	if (Pressure > 0)
		BlastState = TOSMOKE;
	else
		BlastState = NONE;
}


// smoke should be running only in atmosphere or on Earth surface, not on the Moon!
void R7_wreck::StartSmoke()
{
	// create a smoke cloud from the wreck
	PARTICLESTREAMSPEC BlastSmoke = {
		0,			// flags, reserved
		0.5,		// srcsize, particle size at creation, m
		10,			// srcrate, average particle generation rate, Hz   
		0,			// V0, average particle emission velocity, m/s
		0,			// srcspread, emission velocity distribution randomisation
		2,			// lifetime, average particle lifetime [s]
		10,			// growthrate, particle growth rate [m/s]
		5,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 
		0, 
		0.5,
		PARTICLESTREAMSPEC::ATM_FLAT, 
		1.22, 
		1.25
	};

	AddParticleStream (&BlastSmoke, _V(0,0,0), _V(0,1,0), &SmokeLevel);
	SmokeLevel = 1;
	BlastState = SMOKING;
}


void R7_wreck::StartDust()
{
	PARTICLESTREAMSPEC Dust = {
		0,			// flags, reserved
		0.1,			// srcsize, particle size at creation, m
		20,			// srcrate, average particle generation rate, Hz   
		1,			// V0, average particle emission velocity, m/s
		0,			// srcspread, emission velocity distribution randomisation
		4,			// lifetime, average particle lifetime [s]
		3,			// growthrate, particle growth rate [m/s]
		5,			// atmslowdown, deceleration rate b in atmosphere, defined as v = v0 e- bdt 
		PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 
		0, 
		0.5,
		PARTICLESTREAMSPEC::ATM_FLAT, 
		1.22, 
		1.25
	};

	AddParticleStream (&Dust, _V(0,0,0), _V(0,1,0), &DustLevel);
	DustLevel = 1;
	BlastState = DUSTING;

	// change flags to stop the blast after one second
	StopDustTime = oapiGetSimTime() + 0.2;
}



void R7_wreck::StopDust()
{
	DustLevel = 0;
	BlastState = NONE;
}



void R7_wreck::clbkVisualCreated (VISHANDLE vis,int refcount)
{
	VisHandle = vis;
	HideShadows();
}


void R7_wreck::HideShadows()
{
	if (!bLoadMesh)
		return;

	hCraterMesh = GetMesh(VisHandle,0);
	if (hCraterMesh == NULL) 
		return;

	// kill shadows from the craters
	unsigned int i;
	for (i=0; i<oapiMeshGroupCount(hCraterMesh); i++)
	{
		MESHGROUP *mg = oapiMeshGroup(hCraterMesh, i);
		mg->UsrFlag=1; 
	}

	// cleanup
	hCraterMesh = NULL;
}


		// SOUND
void R7_wreck::clbkPostCreation (void)
{
}


// Vessel dll callbacks

DLLCLBK VESSEL *ovcInit (OBJHANDLE hVessel, int flightmodel)
{
	return new R7_wreck (hVessel,flightmodel);
}

DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) 
		delete (R7_wreck*)vessel;
}









